# coding=utf-8
from django.core.management import BaseCommand
from moneta.repository.models import Element

__author__ = 'flanker'


class Command(BaseCommand):
    def handle(self, *args, **options):
        for elt in Element.objects.all():
            print('filename: %s, archive: %s, name: %s, version: %s' %
                  (elt.filename, elt.archive, elt.name, elt.version))


# -*- coding: utf-8 -*-
import pkg_resources

from moneta.repositories.tests import RepositoryTestCase
from moneta.repositories.yum import Yum

__author__ = 'flanker'


class TestYumRpm(RepositoryTestCase):

    def test_add_file(self):
        repo = self.create_repository(Yum)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', '389-ds-base-libs-1.3.3.1-13.el7.x86_64.rpm')
        self.add_file_to_repository(repo, filename)

    def test_generate_index(self):
        repo = self.create_repository(Yum)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', '389-ds-base-libs-1.3.3.1-13.el7.x86_64.rpm')
        self.add_file_to_repository(repo, filename)
        yum = Yum()
        yum.generate_indexes(repo)



# -*- coding: utf-8 -*-
import os
from django.contrib.auth.models import User
from django.core.files.uploadedfile import UploadedFile
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from moneta.repositories.base import RepositoryModel
from moneta.repository.models import Repository, ArchiveState
from moneta.views import generic_add_element

__author__ = 'flanker'


class RepositoryTestCase(TestCase):

    def get_request(self):
        user = User.objects.get_or_create(username='test_user')[0]
        request = HttpRequest()
        request.user = user
        return request

    def create_repository(self, repo_cls, name='test_repo', author=None, states='qualif prod'):
        assert issubclass(repo_cls, RepositoryModel)
        if author is None:
            author = User.objects.get_or_create(username='test_user')[0]
        repo = Repository(author=author, name=name, on_index=True, archive_type=repo_cls.archive_type, is_private=False)
        repo.save()
        for state in states.split():
            ArchiveState(repository=repo, name=state, author=author).save()
        return repo

    def add_file_to_repository(self, repo, filename, states=None, name='', archive='', version=''):
        if states is None:
            states = ['qualif', 'prod', ]
        uploaded_file = UploadedFile(name=os.path.basename(filename), file=open(filename, 'rb'))
        request = self.get_request()
        element = generic_add_element(request, repo, uploaded_file, states, name=name, archive=archive, version=version, )
        return element



from django.conf import settings
from django.http import HttpResponseBadRequest
from django.utils.translation import ugettext_lazy as _
from django.conf.urls import patterns
from moneta.repository.models import Element
from moneta.utils import import_path

__author__ = 'flanker'


class RepositoryModelsClasses(object):
    _models = None

    @classmethod
    def get_model(cls, cls_name):
        return cls.get_models()[cls_name]

    @classmethod
    def get_models(cls):
        if cls._models is None:
            # noinspection PyPep8Naming
            cls._models = {}
            for mw_path in settings.REPOSITORY_CLASSES:
                modelcls = import_path(mw_path)()
                cls._models[modelcls.archive_type] = modelcls
        return cls._models

    def __iter__(self):
        for x, y in self.get_models().items():
            yield x, str(y)


class RepositoryModel(object):
    verbose_name = _('Repository')
    archive_type = 'RepositoryModel'

    def __str__(self):
        """
        returns a unicode representation of this repository
        :return: unicode representation
        """
        return self.verbose_name

    def is_file_valid(self, uploaded_file):
        """ Evaluate if the uploaded file is valid for this repository model
        :param uploaded_file:
        :return: True if this element is acceptable for this model, else False
        :raise:
        """
        return True

    def delete_element(self, rel):
        pass

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data
        """
        return ''

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        Currently, only add all states to this element.
        :param element: Element
        :param states: list of ArchiveState
        """
        element.states.add(*states)

    def url_list(self):
        """
        Return a list of URL patterns specific to this repository
        :return: a patterns as expected by django
        """
        return []

    def public_url_list(self):
        return []

    # noinspection PyMethodMayBeStatic
    def get_index_view(self, repo):
        return repo.model + ':index'

    def wrap_view(self, view, csrf_exempt=False):
        def wrapper(request, *args, **kwargs):
            try:
                return getattr(self, view)(request, *args, **kwargs)
            except AttributeError:
                return HttpResponseBadRequest()
        if csrf_exempt:
            wrapper.csrf_exempt = True
        return wrapper

    def extra_index_urls(self, request, rid):
        """ Provides an iterable of tuples (URL, name), to add on the main index
        :param request: HttpRequest
        :param rid: Repository id
        """
        return []

    @property
    def urls(self):
        """
        Provides URLconf details for the ``Api`` and all registered
        ``Resources`` beneath it.
        """
        pattern_list = self.url_list()
        urlpatterns = patterns('', *pattern_list)
        return urlpatterns

    @property
    def public_urls(self):
        pattern_list = self.public_url_list()
        urlpatterns = patterns('', *pattern_list)
        return urlpatterns



from django.conf import settings
from django.core.signing import Signer, base64_hmac, BadSignature
from django.utils.crypto import constant_time_compare
from django.utils.encoding import force_str, force_text
from django.utils.translation import ugettext as _
import gnupg
import logging

__author__ = 'flanker'
logger = logging.getLogger('django.requests')


class RSASigner(Signer):

    def __init__(self, key=None, sep=':', salt=None):
        super().__init__(sep=sep, salt=salt)
        self.key = str(key or settings.RSA_SECRET_KEY)

    def signature(self, value):
        signature = base64_hmac(self.salt + 'signer', value, self.key)
        # Convert the signature from bytes to str only on Python 3
        return force_str(signature)

    def sign_file(self, fd):
        return self.signature(fd.read())


class DSASigner(RSASigner):

    def __init__(self, key=None, sep=':', salt=None):
        super().__init__(sep=sep, salt=salt)
        self.key = str(key or settings.DSA_SECRET_KEY)

    def signature(self, value):
        signature = base64_hmac(self.salt + 'signer', value, self.key)
        # Convert the signature from bytes to str only on Python 3
        return force_str(signature)

try:
    GPG = gnupg.GPG(gnupghome=settings.GNUPG_HOME, gpgbinary=settings.GNUPG_PATH)

    class GPGSigner(Signer):

        def __init__(self, key=None, sep=':', salt=None):
            super().__init__(sep=sep, salt=salt)
            self.key = str(key or settings.GNUPG_KEYID)

        def signature(self, value):
            return force_str(GPG.sign(value, keyid=self.key, detach=True))

        def sign_file(self, fd):
            # noinspection PyProtectedMember
            return force_str(GPG.sign_file(fd, keyid=self.key, detach=True))

        def export_key(self):
            return GPG.export_keys(self.key)

        def sign(self, value):
            value = force_str(value)
            return str('%s%s%s') % (value, self.sep, self.signature(value))

        def unsign(self, signed_value):
            signed_value = force_str(signed_value)
            if not self.sep in signed_value:
                raise BadSignature('No "%s" found in value' % self.sep)
            value, sig = signed_value.rsplit(self.sep, 1)
            if constant_time_compare(sig, self.signature(value)):
                return force_text(value)
            raise BadSignature('Signature "%s" does not match' % sig)

        # noinspection PyMethodMayBeStatic
        def verify(self, value):
            return GPG.verify(value)

        # noinspection PyMethodMayBeStatic
        def verify_file(self, fd, path_to_data_file=None):
            return GPG.verify_file(fd, path_to_data_file)

except ImportError:
    gnupg = None
    GPG = None
    logger.error(_('unable to import gnugpg module.'))



from distutils.version import LooseVersion
import os.path
import tarfile
import zipfile

from django.conf import settings
from django.conf.urls import url
from django.core.exceptions import PermissionDenied
from django.core.files.uploadedfile import UploadedFile
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.utils.timezone import get_current_timezone
from django.utils.translation import ugettext as _

from moneta.exceptions import InvalidRepositoryException
from moneta.repositories.base import RepositoryModel
from moneta.templatetags.moneta import moneta_url
from moneta.utils import parse_control_data
from moneta.repositories.aptitude import Aptitude
from moneta.repositories.xmlrpc import XMLRPCSite, register_rpc_method
from moneta.repository.models import storage, Repository, Element, ArchiveState

__author__ = 'flanker'

tz = get_current_timezone()
XML_RPC_SITE = XMLRPCSite()


class PyArchive:
    def __init__(self, compressed_file, compression, prefix):
        self.compression = compression
        self.compressed_file = compressed_file
        self.prefix = prefix

    def get_pkg_info(self):
        # noinspection PyBroadException
        try:
            if self.compression == 'egg':
                control_data_file = self.compressed_file.open(os.path.join(self.prefix, 'EGG-INFO', 'PKG-INFO'))
            elif self.compression == 'tar':
                control_data_file = self.compressed_file.extractfile(os.path.join(self.prefix, 'PKG-INFO'))
                # poulating different informations on the element
            elif self.compression == 'whl':
                control_data_file = None
                for name in self.compressed_file.namelist():
                    components = name.split(os.path.sep)
                    if components[0].endswith('.dist-info') and components[1:] == ['METADATA']:
                        control_data_file = self.compressed_file.open(name)
                if control_data_file is None:
                    return None
            else:  # zip
                control_data_file = self.compressed_file.open(os.path.join(self.prefix, 'PKG-INFO'))
            control_data_value = control_data_file.read().decode('utf-8')
            control_data_file.close()
        except:
            return None
        return control_data_value

    def close(self):
        self.compressed_file.close()


class Pypi(Aptitude):
    verbose_name = _('Pypi repository for Python packages')
    storage_uid = '61610000-0000-0000-0000-%012d'
    archive_type = 'pypy'

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        :param element: Element
        :param states: list of ArchiveState
        """
        RepositoryModel.finish_element(self, element, states)

    def is_file_valid(self, uploaded_file: UploadedFile):
        return True

    @staticmethod
    def open_file(filename, archive_file):
        """ identify compression type and return a tuple (open compressed file, compression_type, prefix).
        If compression type is unknown, return None
        """
        endswith = filename.endswith
        compressed_file = None
        # noinspection PyBroadException
        try:
            if endswith('.tar.gz') or endswith('.tar.xz') or endswith('.tar.bz2'):
                compressed_file = tarfile.open(name='file.tar', mode='r:*', fileobj=archive_file)
                prefix = os.path.commonprefix(compressed_file.getnames())
                return PyArchive(compressed_file, 'tar', prefix)
            if endswith('.zip'):
                compressed_file = zipfile.ZipFile(archive_file)
                prefix = os.path.commonprefix(compressed_file.namelist())
                return PyArchive(compressed_file, 'zip', prefix)
            if endswith('.egg'):
                compressed_file = zipfile.ZipFile(archive_file)
                prefix = os.path.commonprefix(compressed_file.namelist())
                return PyArchive(compressed_file, 'egg', prefix)
            if endswith('.whl'):
                compressed_file = zipfile.ZipFile(archive_file)
                prefix = os.path.commonprefix(compressed_file.namelist())
                return PyArchive(compressed_file, 'whl', prefix)
        except Exception:
            if compressed_file:
                compressed_file.close()
        return None

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key)
        py_archive = self.open_file(element.filename, archive_file)
        if py_archive is None:
            raise InvalidRepositoryException(_('Unable to open file'))
        try:
            control_data_value = py_archive.get_pkg_info()
            if not control_data_value:
                raise InvalidRepositoryException(_('No control data in archive'))
            element.extra_data = control_data_value
            control_data = parse_control_data(control_data_value, continue_line='        ', skip_after_blank=True)
            for key, attr in (('Name', 'archive'), ('Version', 'version'), ('Home-page', 'official_link'),
                              ('Description', 'long_description')):
                if key in control_data:
                    setattr(element, attr, control_data.get(key, ''))
            element.archive = element.archive.replace('-', '').replace('_', '')
            element.name = element.archive
        finally:
            py_archive.close()
            archive_file.close()

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            *

        :return: a patterns as expected by django

        """
        pattern_list = [
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/simple/?$",
                self.wrap_view('simple'), name="simple"),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/simple/"
                "(?P<search_pattern>[a-z\d_\-]*)/?$", self.wrap_view('simple'),
                name="simple"),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/rpc/?$",
                self.wrap_view('xmlrpc', csrf_exempt=True), name="xmlrpc"),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/simple/?$", self.wrap_view('simple'), name="simple"),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/simple/(?P<search_pattern>[a-zA-Z\d_\-]*)/?$",
                self.wrap_view('simple'), name="simple"),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/rpc/?$",
                self.wrap_view('xmlrpc', csrf_exempt=True), name="xmlrpc"),
            url(r"^\d+/[\w\-\._]+/[\w\-\._]+/f/(?P<eid>\d+)/.*$", self.wrap_view('get_filename'), name="get_file"),
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def index(self, request, rid):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = list(ArchiveState.objects.filter(repository=repo).order_by('name'))
        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'admin_allowed': repo.admin_allowed(request), }
        view_name = moneta_url(repo, 'simple')
        tab_infos = [
            (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug}), states, ArchiveState(name=_('All states'), slug='all-states')),
        ]
        for state in states:
            tab_infos.append(
                (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'state_slug': state.slug}), [state], state)
            )
        template_values['tab_infos'] = tab_infos
        return render_to_response('repositories/pypi/index.html', template_values, RequestContext(request))

    @staticmethod
    def get_filename(request, eid):
        from moneta.views import get_file

        return get_file(request, eid)

    # noinspection PyUnusedLocal
    def simple(self, request, rid, repo_slug, state_slug=None, search_pattern=''):
        search_pattern = search_pattern.replace('-', '').replace('_', '')
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        if search_pattern:
            base_query = base_query.filter(archive__iexact=search_pattern)
        view_name = moneta_url(repo, 'get_file')
        elements = [(x.filename, x.md5, reverse(view_name, kwargs={'eid': x.id, })) for x in base_query[0:1000]]
        template_values = {'elements': elements, 'rid': rid, }
        return render_to_response('repositories/pypi/simple.html', template_values, RequestContext(request))

    def xmlrpc(self, request, rid, repo_slug, state_slug):
        return XML_RPC_SITE.dispatch(request, self, rid, repo_slug, state_slug)

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='search')
    def xr_search(request, rpc_args, self, rid, repo_slug, state_slug=None):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        state = get_object_or_404(ArchiveState, repository=repo, name=state_slug) if state_slug else None
        global_and = len(rpc_args) == 0 or rpc_args[1] != 'or'
        if len(rpc_args) == 0 or not isinstance(rpc_args[0], dict):
            raise PermissionDenied
        filters = None
        for query_name, attr_name in (('name', 'archive'), ('version', 'version'), ('home_page', 'official_link'),
                                      ('description', 'long_description'),):
            if query_name in rpc_args[0]:
                value = rpc_args[0][query_name]
                if isinstance(value, list):
                    if value:
                        value = value[0]
                    else:
                        value = ''
                value = value.replace('-', '').replace('_', '')
                sub_query = Q(**{attr_name + '__icontains': value})
                if filters is None:
                    filters = sub_query
                elif global_and:
                    filters = filters and sub_query
                else:
                    filters = filters or sub_query
        query = Element.objects.filter(repository=repo)
        if state:
            query = query.filter(states=state)
        if filters is not None:
            query = query.filter(filters)
        res = [{'name': x.archive, 'version': x.version, 'summary': x.long_description,
                '_pypi_ordering': 1.0} for x in query]
        return res

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='list_packages')
    def xr_list_packages(request, rpc_args, self, rid, repo_slug, state_slug=None):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            query = Element.objects.filter(repository=repo, states=state).order_by('name')
        else:
            query = Element.objects.filter(repository=repo).order_by('name')
        results = []
        prev = None
        for value in query:
            if value.name != prev:
                results.append(value.name)
                prev = value.name
        return results

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='package_releases')
    def xr_package_releases(request, rpc_args, self, rid, repo_slug, state_slug=None):
        if len(rpc_args) == 0:
            raise PermissionDenied
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            query = Element.objects.filter(repository=repo, states=state, archive=rpc_args[0])
        else:
            query = Element.objects.filter(repository=repo, archive=rpc_args[0])
        versions = [LooseVersion(x.version) for x in query]
        versions.sort()
        return [str(x) for x in versions]

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='user_packages')
    def xr_user_packages(request, rpc_args, self, rid, repo_slug, state_slug=None):
        if len(rpc_args) == 0:
            raise PermissionDenied
        author = str(rpc_args[0]).lower()
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            query = Element.objects.filter(repository=repo, state=state).filter(author__name__icontains=author) \
                .order_by('name')
        else:
            query = Element.objects.filter(repository=repo).filter(author__name__icontains=author).order_by('name')
        results = []
        prev = None
        for value in query:
            if value.name != prev:
                results.append(('Owner', value.name))
                prev = value.name
        return results

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='release_urls')
    def xr_release_urls(request, rpc_args, self, rid, repo_slug, state_slug=None):
        if len(rpc_args) != 2:
            raise PermissionDenied
        name = rpc_args[0]
        version = rpc_args[1]
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            query = Element.objects.filter(repository=repo, state=state).filter(archive=name, version=version)
        else:
            query = Element.objects.filter(repository=repo).filter(archive=name, version=version)
        results = []
        # noinspection PyUnresolvedReferences
        base = request.build_absolute_uri('/')[:-1]
        for element in query:
            view_name = moneta_url(repo, 'get_file')
            package_url = base + reverse(view_name, kwargs={'eid': element.id, })
            results.append({'url': package_url, 'packagetype': 'sdist',
                            'filename': element.filename, 'size': element.filesize, 'downloads': 0,
                            'comment_text': element.long_description, 'md5_dist': element.md5, 'has_sig': False,
                            'python_version': 'source', })
        return results

    # noinspection PyMethodParameters,PyUnusedLocal
    @register_rpc_method(XML_RPC_SITE, name='release_data')
    def xr_release_data(request, rpc_args, self, rid, repo_slug, state_slug=None):
        if len(rpc_args) != 2:
            raise PermissionDenied
        name = rpc_args[0]
        version = rpc_args[1]
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.model)
        if state_slug:
            stat = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            query = list(Element.objects.filter(repository=repo, state=stat).filter(archive=name, version=version)[0:1])
        else:
            query = list(Element.objects.filter(repository=repo).filter(archive=name, version=version)[0:1])
        if not query:
            return {}
        element = query[0]
        result = {}
        for query_name, attr_name in (('name', 'archive'), ('version', 'version'), ('home_page', 'official_link'),
                                      ('description', 'long_description'), ('author', 'author')):
            result[query_name] = str(getattr(element, attr_name))
        for k in ('author_email', 'maintainer', 'maintainer_email', 'license', 'summary', 'keywords', 'platform',
                  'download_url'):
            result[k] = ''
        return result

    # noinspection PyMethodParameters,PyUnusedLocal,PyMethodMayBeStatic
    @register_rpc_method(XML_RPC_SITE, name='package_roles')
    def xr_package_roles(request, rpc_args, self, rid, repo_slug, state_slug=None):
        return []



# -*- coding=utf-8 -*-
"""
Emulate a Maven3-compatible repository.


/$groupId[0]/../${groupId[n]/$artifactId/$version/$artifactId-$version.$extension
/$groupId[0]/../$groupId[n]/$artifactId/$version/$artifactId-$version-$classifier.$extension
/org/codehaus/plexus/plexus-container/0.15/
                     /plexus-container-0.15.jar
                     /plexus-container-0.15.jar.md5
                     /plexus-container-0.15.pom
                     /plexus-container-0.15.pom.md5
                     /plexus-container-0.15-javadoc.jar
                     /plexus-container-0.15-javadoc.jar.md5
                     /plexus-container-0.15-javasrc.jar
                     /plexus-container-0.15-javasrc.jar.md5

Maven3 variables:
    * $groupId = org.codehaus.plexus
    * $artifactId = plexus-container
    * $version = 0.15
    * $extension = jar
    * $classifier = javadoc, javasrc

Moneta mapping:
    * element.version maps to $version
    * element.archive maps to $groupId.$artifactId
    * element.name maps to $artifactId => built from element.archive
    * element.filename maps to $artifactId-$version-$classifier.$extension
    * element.full_name = element.filename

Command for uploading files:
curl --data-binary @$FILENAME http://[...]/\?filename=$FILENAME\&states=prod\&states=qualif\&archive=$groupId.$artifactId\&version=$version

URLs to emulate:
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename)$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).md5$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).sha1$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).sha256$
"""
import os.path
import zipfile

from django.conf import settings
from django.conf.urls import url
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.template.loader import render_to_string

from django.utils.translation import ugettext as _
from moneta.repositories.base import RepositoryModel
from moneta.templatetags.moneta import moneta_url

from moneta.utils import parse_control_data
from moneta.repositories.aptitude import Aptitude
from moneta.repository.models import storage, Repository, Element, ArchiveState
from moneta.views import get_file

__author__ = 'flanker'


class Maven3(Aptitude):
    verbose_name = _('Maven repository for Java packages')
    storage_uid = '94633000-0000-0000-0000-%012d'
    archive_type = 'maven3'

    def is_file_valid(self, uploaded_file):
        return True

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        if element.archive:
            element.name = element.archive.rpartition('.')[2]
        if element.filename.endswith('.jar') and False:
            archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key)

            compressed_file = zipfile.ZipFile(archive_file)
            prefix = os.path.commonprefix(compressed_file.namelist())
            control_data_file = compressed_file.open(os.path.join(prefix, 'META-INF', 'MANIFEST.MF'))

            control_data_value = control_data_file.read().decode('utf-8')
            control_data_file.close()
            compressed_file.close()
            archive_file.close()
            element.extra_data = control_data_value
            control_data = parse_control_data(control_data_value, continue_line=' ')
            for key, attr in (('Bundle-SymbolicName', 'name'), ('Bundle-Version', 'version'),
                              ('Implementation-Title', 'archive'), ('Implementation-Version', 'version'),
                              ('Name', 'name'),):
                if key in control_data:  # archive : PackageName, name : Organization Name
                    setattr(element, attr, control_data.get(key, ''))
                    # element.filename = '%s-%s.jar' % (element.archive.rpartition('.')[2], element.version)

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        :param element: Element
        :param states: list of ArchiveState
        """
        RepositoryModel.finish_element(self, element, states)

    @staticmethod
    def browse_repo_inner(rid: int, query_string: str, state_slug: str=None):
        """should only called from browse_repository()

        :param rid:
        :param query_string: a single str corresponding to $groupId/$artifactId/$version/$filename.$extension
        :return:
            a string -> send it in a HttpResponse
            an Element -> send it with views.get_file
            a dict -> {"net": {"19pouces": {"moneta": {"1.6.0": {set of <Element>}, }, }, }
        """
        if query_string[-1:] == '/':
            query_string = query_string[:-1]
        state = get_object_or_404(ArchiveState, repository__id=rid, name=state_slug) if state_slug else None
        gav, sep, filename = query_string.rpartition('/')
        dotted_gavf = query_string.replace('/', '.')
        if sep == '/':
            group_artifact, sep, version = gav.rpartition('/')
            group_artifact = group_artifact.replace('/', '.')
            dotted_gav = gav.replace('/', '.')
            if sep == '/':
                # URL: "gro/up/arti/fact/version/filename*"
                # URL: "gro/up/arti/fact/version*"
                # URL: "gro/up/arti/fact*"
                no_ext, sep, ext = filename.rpartition('.')
                if ext not in ('sha1', 'sha256', 'md5') and sep == '.':
                    no_ext = filename
                query = Element.objects.filter(
                    Q(archive=group_artifact, version=version, filename=no_ext) |
                    Q(archive=dotted_gav, version=filename) |
                    Q(archive=dotted_gavf) | Q(archive__startswith=dotted_gavf + '.')
                )
            else:
                # URL: "group_artifact/version"
                query = Element.objects.filter(
                    Q(archive=dotted_gav, version=filename) |
                    Q(archive=dotted_gavf) | Q(archive__startswith=dotted_gavf + '.')
                )
        else:
            # URL: "group_artifact"
            query = Element.objects.filter(archive__startswith=dotted_gavf)
        query = query.filter(repository__id=rid)
        if state:
            query = query.filter(states=state)
        no_ext, sep, ext = query_string.rpartition('.')
        if ext not in ('sha1', 'sha256', 'md5') and sep == '.':
            no_ext = query_string

        # ok, we have to check
        results = {}
        for elt in query:
            elt_gavf = '%s/%s/%s' % (elt.archive.replace('.', '/'), elt.version, elt.filename)

            if elt_gavf == no_ext:
                if ext in ('sha1', 'sha256', 'md5'):
                    return getattr(elt, ext)
                return elt
            elt_gav = '%s/%s' % (elt.archive.replace('.', '/'), elt.version)
            if elt_gav == query_string:
                pass
            parents_results = results
            for path_comp in elt.archive.split('.'):
                parents_results = parents_results.setdefault(path_comp, {})
            parents_results.setdefault(elt.version, set()).add(elt)
        return results

    def browse_repository(self, request: HttpRequest, rid: int, repo_slug: str, query_string: str='', state_slug: str=None) -> HttpResponse:
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        result = self.browse_repo_inner(rid, query_string, state_slug=state_slug)
        if isinstance(result, str):  # sha1/sha256/md5
            return HttpResponse(result)
        elif isinstance(result, Element):
            return get_file(request, eid=result.pk, element=result)
        assert isinstance(result, dict)

        new_query_string = ''
        bread_crumbs = [(_('Root'), self.get_browse_url(repo, new_query_string, state_slug=state_slug))]
        while len(result) == 1 and isinstance(result, dict):
            path_component, result = result.popitem()
            new_query_string += path_component + '/'
            bread_crumbs.append((path_component, self.get_browse_url(repo, new_query_string, state_slug=state_slug)))
        if isinstance(result, set):
            url_list = []
            for elt in result:
                new_gavf_elt_filename = new_query_string + elt.filename
                elt_url = self.get_browse_url(repo, new_gavf_elt_filename, state_slug=state_slug)
                url_list += [(new_gavf_elt_filename, elt_url, elt.id),
                             (new_gavf_elt_filename + '.sha1', elt_url + '.sha1', None),
                             (new_gavf_elt_filename + '.sha256', elt_url + '.sha256', None),
                             (new_gavf_elt_filename + '.md5', elt_url + '.md5', None),
                             ]
        else:
            assert isinstance(result, dict)
            url_list = [(new_query_string + key, self.get_browse_url(repo, new_query_string + key, state_slug=state_slug), None)
                        for key in result]
        template_values = {'repo': repo, 'upload_allowed': repo.upload_allowed(request), 'repo_slug': repo_slug, 'admin': True,
                           'paths': url_list, 'request_path': new_query_string, 'state_slug': state_slug,
                           'bread_crumbs': bread_crumbs, }
        status_code = 200 if url_list else 404
        return render_to_response('repositories/maven3/browse.html', template_values, RequestContext(request), status=status_code)

    @staticmethod
    def get_browse_url(repo, new_query_string, state_slug=None):
        browse_view_name = moneta_url(repo, 'browse')
        if state_slug is None:
            return reverse(browse_view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'query_string': new_query_string})
        return reverse(browse_view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'query_string': new_query_string, 'state_slug': state_slug, })

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename)(|.md5|sha1|sha256)$
            r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/(?P<gav>.+)/(?P<filename>[^/]+)$'
            '
        :return: a patterns as expected by django

        """

        # http://mvnrepository.com/artifact/org.requs/requs-exec/1.11
        pattern_list = [
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/browse/(?P<query_string>.*)$', self.wrap_view('browse_repository'), name='browse'),
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/s/(?P<state_slug>[\w\-\._]+)/browse/(?P<query_string>.*)$', self.wrap_view('browse_repository'), name='browse'),
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/browse/$', self.wrap_view('browse_repository'), name='browse'),
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/s/(?P<state_slug>[\w\-\._]+)/browse/$', self.wrap_view('browse_repository'), name='browse'),
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def index(self, request, rid):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = list(ArchiveState.objects.filter(repository=repo).order_by('name'))
        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'admin_allowed': repo.admin_allowed(request), }
        state_infos = []
        template_values['state_slug'] = None
        request_context = RequestContext(request)
        setting_str = render_to_string('repositories/maven3/maven_settings.xml', template_values, request_context)
        state_infos.append(('all-packages', str(setting_str), _('All states'), states))
        for state in states:
            template_values['state_slug'] = state.slug
            setting_str = render_to_string('repositories/maven3/maven_settings.xml', template_values, request_context)
            state_infos.append((state.slug, str(setting_str), state.name, [state]))
        template_values['state_infos'] = state_infos
        return render_to_response('repositories/maven3/index.html', template_values, request_context)



# coding=utf-8
from xmlrpc.client import dumps, Fault, loads
from django.http.response import HttpResponse


class XMLRPCSite(object):
    def __init__(self):
        self.methods = {}

    def register_method(self, func, name=None):
        if name is None:
            name = func.__name__
        self.methods[name] = func

    def dispatch(self, request, *args, **kwargs):
        rpc_call = loads(request.body.decode('utf-8'))
        rpc_args = rpc_call[0]
        method_name = rpc_call[1]
        try:
            # noinspection PyCallingNonCallable
            src_result = self.methods[method_name](request, rpc_args, *args, **kwargs)
            result = src_result,
        except Exception as e:
            result = Fault(e.__class__.__name__, str(e))
        data = dumps(result, method_name, True)
        return HttpResponse(data, content_type='application/xml+rpc')


def register_rpc_method(site, name=None):
    def decorated(func):
        site.register_method(func, name=name)
        return func
    return decorated



# coding=utf-8
# noinspection PyCompatibility
import bz2
import gzip
import hashlib
import mimetypes
import os
import tarfile
import zipfile

from moneta.archives import ArFile
from moneta.utils import mkdtemp


__author__ = 'flanker'

mimetypes.init()


# noinspection PyUnusedLocal
def informations(element, open_file, filename, temp_files, uncompressed_path=None):
    size = 0
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()
    data = open_file.read(1024)
    while data:
        md5.update(data)
        sha1.update(data)
        sha256.update(data)
        size += len(data)
        data = open_file.read(1024)
    element.sha1 = sha1.hexdigest()
    element.sha256 = sha256.hexdigest()
    element.md5 = md5.hexdigest()
    element.filesize = size
    element.extension = os.path.splitext(filename)[1]
    element.mimetype = mimetypes.guess_type(filename, strict=False)[0]
    if element.mimetype is None:
        element.mimetype = 'application/octet-stream'
    element.filename = filename


def compressions(element, open_file, filename, temp_files, uncompressed_path=None):
    if uncompressed_path is not None:
        return None
    extension = os.path.splitext(filename)[1]
    new_filename = filename[0:-len(extension)]
    result_file = None
    if extension == '.zip':
        result_file = mkdtemp()  # create a new temp directory
        temp_files.add(result_file)  # ensure that all extracted files are eventually deleted
        obj = zipfile.ZipFile(open_file, 'r')
        members = filter(lambda x: os.path.abspath(os.path.join(result_file, x)).find(result_file) == 0, obj.namelist())
        # filter members that should be created outside temp_dir (e.g. names begin with /)
        obj.extractall(result_file, members)
        obj.close()
    elif extension == '.gz':
        result_file = mkdtemp()
        temp_files.add(result_file)
        result_file = os.path.join(result_file, new_filename)
        with gzip.GzipFile(filename, 'rb', 9, open_file) as obj:
            with open(result_file, 'wb') as fd:
                data = obj.read(4096)
                while data != '':
                    fd.write(data)
                    data = obj.read(4096)
        with open(result_file, 'rb') as fd:
            extra_compression = compressions(element, fd, new_filename, temp_files)
        if extra_compression is not None:
            result_file = extra_compression
    elif extension == '.bz2':
        result_file = mkdtemp()
        temp_files.add(result_file)
        result_file = os.path.join(result_file, new_filename)
        decompressor = bz2.BZ2Decompressor()
        with open(result_file, 'wb') as fd:
            data = open_file.read(4096)
            while data != '':
                data = decompressor.decompress(data)
                fd.write(data)
                data = open_file.read(4096)
        with open(result_file, 'rb') as fd:
            extra_compression = compressions(element, fd, new_filename, temp_files)
        if extra_compression is not None:
            result_file = extra_compression
    elif extension in ('.tgz', '.tbz', '.tar'):
        result_file = mkdtemp()  # create a new temp directory
        temp_files.add(result_file)  # ensure that all extracted files are eventually deleted
        obj = tarfile.TarFile(fileobj=open_file, mode='r')
        members = filter(lambda x: x.type in (tarfile.REGTYPE, tarfile.DIRTYPE) and
                         os.path.abspath(os.path.join(result_file, x.name)).find(result_file) == 0,
                         obj.getmembers())
        # filter members that should be created outside temp_dir (e.g. names begin with /)
        obj.extractall(result_file, members)
        obj.close()
    if result_file and os.path.isdir(result_file):
        if len(os.listdir(result_file)[0:2]) == 1:
            result_file = os.path.join(result_file, os.listdir(result_file)[0])
        else:
            tmp_dir = mkdtemp()
            temp_files.add(tmp_dir)
            tmp_dir = os.path.join(tmp_dir, new_filename)
            os.makedirs(tmp_dir)
            for name in os.listdir(result_file):
                os.rename(os.path.join(result_file, name), os.path.join(tmp_dir, name))
            result_file = tmp_dir
    return result_file


# noinspection PyUnusedLocal
def deb_archive(element, open_file, filename, temp_files, uncompressed_path=None):
    """
    Uncompress files in a Debian .deb archive.
    A .deb is a .ar archive with control.tar.gz (metadata) and data.tar.gz (actual package files).
    :param element:
    :param open_file:
    :param filename:
    :param temp_files:
    :param uncompressed_path:
    :return:
    """
    if uncompressed_path is not None:  # another filter has uncompressed this file
        return None
    extension = os.path.splitext(filename)[1]
    if extension != '.deb':
        return None
    new_filename = filename[0:-len(extension)]
    ar_file_obj = ArFile(name=new_filename, mode='r', fileobj=open_file)  # allow to extract files from the .deb
    data_file_obj = ar_file_obj.extractfile('data.tar.gz') or ar_file_obj.extractfile('data.tar.gz/')
    # file descriptor to the data.tar.gz file
    if data_file_obj is None:
        return None
    result_file = mkdtemp()  # create a new temp directory
    temp_files.add(result_file)  # ensure that all extracted files are eventually deleted
    tar_file_obj = tarfile.open(name='data.tar.gz', fileobj=data_file_obj, mode='r:gz')
    # allow to extract files from the data.tar.gz from the .deb file
    # all these operations use the same original file descriptor (no need to really extract the data.tar.gz file)
    members = filter(lambda x: x.type in (tarfile.REGTYPE, tarfile.DIRTYPE) and
                     os.path.abspath(os.path.join(result_file, x.name)).find(result_file) == 0,
                     tar_file_obj.getmembers())
    # filter members that should be created outside temp_dir (e.g. names begin with /)
    tar_file_obj.extractall(result_file, members)
    tar_file_obj.close()
    data_file_obj.close()
    ar_file_obj.close()
    if result_file and os.path.isdir(result_file):
        if len(os.listdir(result_file)[0:2]) == 1:
            result_file = os.path.join(result_file, os.listdir(result_file)[0])
        else:
            tmp_dir = mkdtemp()
            temp_files.add(tmp_dir)
            tmp_dir = os.path.join(tmp_dir, new_filename)
            os.makedirs(tmp_dir)
            for name in os.listdir(result_file):
                os.rename(os.path.join(result_file, name), os.path.join(tmp_dir, name))
            result_file = tmp_dir
    return result_file


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding=utf-8 -*-
from moneta.views import public_check, index, check, add_element, add_element_post, add_element_signature, \
    get_signature, modify_repository, delete_repository, delete_element, search_package, compare_states, private_check, \
    test_upload, get_file_p, get_checksum_p, get_signature_p, get_file, get_checksum, show_file

__author__ = 'flanker'

from django.conf.urls import url

urlpatterns = [
    url(r'^p/$', public_check, name='public_check'),
    url(r'^p/get/(?P<eid>\d+)/(?P<name>.*)$', get_file_p, name='get_file_p'),
    url(r'^p/check/(?P<eid>\d+)/(?P<value>sha1|sha256|md5)$', get_checksum_p, name='get_checksum_p'),
    url(r'^p/signature/(?P<eid>\d+)/(?P<sid>\d+)/$', get_signature_p, name='get_signature_p'),
    url(r'^a/get/(?P<eid>\d+)/(?P<name>.*)$', get_file, name='get_file'),
    url(r'^a/check/(?P<eid>\d+)/(?P<value>sha1|sha256|md5)$', get_checksum, name='get_checksum'),
    url(r'^a/file/(?P<eid>\d+)/$', show_file, name='show_file'),
    url(r'^a/add_package/(?P<rid>\d+)/$', add_element, name='add_element'),
    url(r'^a/post/(?P<rid>\d+)/$', add_element_post, name='add_element_post'),
    url(r'^a/sign/(?P<rid>\d+)/$', add_element_signature, name='add_element_signature'),
    url(r'^a/signature/(?P<eid>\d+)/(?P<sid>\d+)/$', get_signature, name='get_signature'),
    url(r'^a/modify/(?P<rid>\d+)/$', modify_repository, name='modify_repository'),
    url(r'^a/delete/(?P<rid>\d+)/$', delete_repository, name='delete_repository'),
    url(r'^a/delete/(?P<rid>\d+)/(?P<eid>\d+)/$', delete_element, name='delete_element'),
    url(r'^a/search/(?P<rid>\d+)/$', search_package, name='search_package'),
    url(r'^a/compare/(?P<rid>\d+)/$', compare_states, name='compare_states'),
    url(r'^a/index/$', index, name='index'),
    url(r'^a/$', private_check, name='private_check'),
    url(r'^a/check/$', check, name='check'),
    url(r'^test/$', test_upload, name='test_upload'),

]



# -*- coding: utf-8 -*-
import base64
from distutils.version import LooseVersion
import gzip
import json
import subprocess
import tarfile
import io
import tempfile
import zlib

from django.conf import settings
from django.conf.urls import url
from django.core.files.uploadedfile import UploadedFile
from django.core.urlresolvers import reverse
from django.http import HttpResponse, Http404
from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext
from django.utils.translation import ugettext as _
from rubymarshal.classes import UsrMarshal
from rubymarshal.writer import write
from yaml import MappingNode
import yaml
from yaml.composer import Composer
from yaml.constructor import Constructor
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner

from moneta.repositories.aptitude import Aptitude
from moneta.repositories.base import RepositoryModel
from moneta.repository.models import ArchiveState, Element, Repository, storage
from moneta.templatetags.moneta import moneta_url
from moneta.views import sendpath

__author__ = 'flanker'


class RubyMarshal(UsrMarshal):
    def __init__(self, values):
        super().__init__('Gem::' + self.__class__.__name__, values=values)


class Version(RubyMarshal):
    @property
    def version(self):
        return self.values['version']


class Specification(RubyMarshal):
    @property
    def name(self):
        return self.values['name']

    @property
    def version(self) -> Version:
        return self.values['version']

    @property
    def platform(self):
        return self.values['platform']

    @property
    def description(self):
        return self.values['description']

    @property
    def rubygems_version(self):
        return self.values['rubygems_version']

    @property
    def summary(self):
        return self.values['summary']


class Dependency(RubyMarshal):
    @property
    def name(self):
        return self.values['name']

    @property
    def requirement(self):
        return self.values['requirement']

    @property
    def type(self):
        return self.values['type']

    @property
    def prerelease(self):
        return self.values['prerelease']

    @property
    def version_requirements(self):
        return self.values['version_requirements']


class Requirement(RubyMarshal):
    @property
    def requirements(self):
        return self.values['requirements']


class RubyConstructor(Constructor):
    def construct_version(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Version(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_specification(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Specification(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_dependency(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Dependency(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_requirement(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Requirement(super(RubyConstructor, self).construct_mapping(node, deep=deep))


RubyConstructor.add_constructor('!ruby/object:Gem::Specification', RubyConstructor.construct_specification)
RubyConstructor.add_constructor('!ruby/object:Gem::Version', RubyConstructor.construct_version)
RubyConstructor.add_constructor('!ruby/object:Gem::Dependency', RubyConstructor.construct_dependency)
RubyConstructor.add_constructor('!ruby/object:Gem::Requirement', RubyConstructor.construct_requirement)


class RubyLoader(Reader, Scanner, Parser, Composer, RubyConstructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        RubyConstructor.__init__(self)
        Resolver.__init__(self)


class RubyGem(Aptitude):
    verbose_name = _('Gem repository for Ruby packages')
    storage_uid = '71610000-0000-0000-0000-%012d'
    archive_type = 'rubygem'

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        :param element: Element
        :param states: list of ArchiveState
        """
        RepositoryModel.finish_element(self, element, states)
        # self.generate_indexes(element.repository)

    def is_file_valid(self, uploaded_file: UploadedFile):
        if not uploaded_file.name.endswith('.gem'):
            return False
        try:
            fileobj = uploaded_file.file
            fd = tarfile.open(fileobj=fileobj, mode='r|')
            names = fd.getnames()
            fd.close()
        except tarfile.TarError:
            return False
        if 'data.tar.gz' not in names or 'metadata.gz' not in names:
            return False
        return True

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key, '')
        gem_fd = tarfile.open(fileobj=archive_file, mode='r')
        metadata_fd = gem_fd.extractfile('metadata.gz')
        metadata_gz_content = metadata_fd.read()
        metadata_bytes = gzip.decompress(metadata_gz_content)
        gem_fd.close()
        data = yaml.load(io.BytesIO(metadata_bytes), Loader=RubyLoader)
        for key, attr in (('name', 'archive'), ('homepage', 'official_link'), ('summary', 'long_description'),
                          ('name', 'name')):
            if key in data.values:
                setattr(element, attr, data.values[key])
        element.version = data.values['version'].version
        p = subprocess.Popen(['ruby', '-e', 'puts Marshal.dump (Gem::Specification.from_yaml(ARGF.read))'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        stdout, stderr = p.communicate(metadata_bytes)
        extra_data = {'yaml': metadata_bytes.decode('utf-8'), 'marshal': base64.b64encode(stdout).decode('utf-8')}
        element.extra_data = json.dumps(extra_data)

    @staticmethod
    def marshal_data(elt: Element) -> bytes:
        marshal = json.loads(elt.extra_data)['marshal']
        return base64.b64decode(marshal.encode('utf-8'))

    @staticmethod
    def python_data(elt: Element) -> Specification:
        data = json.loads(elt.extra_data)['yaml']
        return yaml.load(io.BytesIO(data.encode('utf-8')), Loader=RubyLoader)

    @staticmethod
    def yaml_data(elt: Element) -> Specification:
        return json.loads(elt.extra_data)['yaml']

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            *

        :return: a patterns as expected by django

        """
        src_pattern_list = [(r'(?P<filename>(specs\.4\.8|prerelease_specs\.4\.8|latest_specs\.4\.8|Marshal\.4\.8|versions\.list|names\.list)(\.gz)?)', 'specs', 'specs'),
                            (r'gems/(?P<filename>.+)', 'download', 'download'),
                            (r'specs/(?P<filename>.+)\.gemspec', 'gem_specs', 'gem_specs'),
                            (r'quick/Marshal\.4\.8/(?P<filename>.+)\.gemspec(?P<compression>(\.rz|))', 'quick_gem_specs', 'quick_gem_specs'),
                            (r'', 'index', 'index'),
                            ]
        pattern_list = []
        for pattern, view, name in src_pattern_list:
            pattern_list.append(
                url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/a/%s$" % pattern, self.wrap_view(view), name=name)
            )
            pattern_list.append(
                url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/%s$" % pattern, self.wrap_view(view), name=name)
            )
        pattern_list += [
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def specs(self, request, rid, repo_slug, state_slug=None, filename='specs.4.8.gz'):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        if state_slug:
            filename = 'specs/%(slug)s/%(filename)s' % {'slug': state_slug, 'filename': filename, }
        else:
            filename = 'specs/%(filename)s' % {'filename': filename, }
        uid = self.storage_uid % repo.pk
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        return sendpath(settings.STORAGE_CACHE, key, filename, 'application/gzip')

    def quick_gem_specs(self, request, rid, repo_slug, state_slug=None, filename=None, compression=''):
        name, sep, version = filename.rpartition('-')
        if sep != '-':
            raise Http404
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, name=name, version=version)
        content = self.marshal_data(element)
        if compression == '.rz':
            content = zlib.compress(content)
        return HttpResponse(content, content_type='application/gzip')

    def gem_specs(self, request, rid, repo_slug, state_slug=None, filename=None):
        name, sep, version = filename.rpartition('-')
        if sep != '-':
            raise Http404
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, name=name, version=version)
        return HttpResponse(self.yaml_data(element), content_type='text/yaml')

    def download(self, request, rid, repo_slug, state_slug=None, filename=None):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, filename=filename)
        from moneta.views import get_file
        return get_file(request, element.pk)

    def index(self, request, rid, repo_slug=None, state_slug=None):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = list(ArchiveState.objects.filter(repository=repo).order_by('name'))
        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'admin_allowed': repo.admin_allowed(request), }
        view_name = moneta_url(repo, 'index')
        tab_infos = [
            (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug}), states,
             ArchiveState(name=_('All states'), slug='all-states')),
        ]
        for state in states:
            tab_infos.append(
                (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'state_slug': state.slug}), [state], state)
            )
        template_values['tab_infos'] = tab_infos
        return render_to_response('repositories/ruby/index.html', template_values, RequestContext(request))

    def generate_indexes(self, repo: Repository, states=None, validity=365):
        if states is None:
            states = list(ArchiveState.objects.filter(repository=repo))
        base_query = Element.objects.filter(repository=repo).prefetch_related('states')
        state_infos = [(None, None)] + [(state.slug, state.pk) for state in states]
        all_elements = {state_info[0]: [] for state_info in state_infos}
        all_specs = {state_info[0]: [] for state_info in state_infos}
        last_elements = {state_info[0]: {} for state_info in state_infos}
        for element in base_query:
            element_spec = self.python_data(element)
            assert isinstance(element_spec, Specification)
            element_data = [element.archive, element_spec.version, element_spec.platform]
            all_specs[None].append(element_spec)
            if 'date' in element_spec.values:
                del element_spec.values['date']
            all_elements[None].append(element_data)
            last_elements[None].setdefault(element_spec.name, []).append((element_spec.version.version, element_spec.platform))
            for state in element.states.all():
                all_elements[state.slug].append(element_data)
                all_specs[state.slug].append(element_spec)
                last_elements[state.slug].setdefault(element_spec.name, []).append((element_spec.version.version, element_spec.platform))

        for state_info in state_infos:
            folder_name = 'specs/' if state_info[0] is None else 'specs/%(slug)s' % {'slug': state_info[0]}

            last_elements_by_state = []
            for name, versions in last_elements[state_info[0]].items():
                try:
                    versions.sort(key=lambda x: LooseVersion(x[0]))
                    last_elements_by_state.append([name, Version([versions[-1][0]]), versions[-1][1].encode('utf-8')])
                except ValueError:
                    continue
            names = list(last_elements[state_info[0]].keys())
            names.sort()

            def versions_write_fn(file_obj):
                for y in names:
                    file_obj.write(('%s %s\n' % (y, ','.join([z[0] for z in last_elements[state_info[0]][y]]))).encode('utf-8'))

            def names_write_fn(file_obj):
                for y in names:
                    file_obj.write(('%s\n' % y).encode('utf-8'))

            self.__write_file(repo, '%s/specs.4.8' % folder_name, lambda x: write(x, all_elements))
            self.__write_file(repo, '%s/latest_specs.4.8' % folder_name, lambda x: write(x, last_elements_by_state))
            self.__write_file(repo, '%s/prerelease_specs.4.8' % folder_name, lambda x: write(x, []))
            self.__write_file(repo, '%s/Marshal.4.8' % folder_name, lambda x: write(x, all_specs))
            self.__write_file(repo, '%s/versions.list' % folder_name, versions_write_fn)
            self.__write_file(repo, '%s/names.list' % folder_name, names_write_fn)

    def __write_file(self, repo, dest_filename, write_function):
        uid = self.storage_uid % repo.pk
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        plain_file = tempfile.NamedTemporaryFile(mode='w+b', dir=settings.TEMP_ROOT, delete=False)
        gz_plain_file = tempfile.NamedTemporaryFile(mode='w+b', dir=settings.TEMP_ROOT, delete=False)
        gz_file = gzip.open(gz_plain_file, 'wb')
        write_function(plain_file)
        plain_file.flush()
        plain_file.seek(0)
        for block in iter(lambda: plain_file.read(8192), b''):
            gz_file.write(block)
        gz_file.close()
        gz_plain_file.close()
        storage(settings.STORAGE_CACHE).import_filename(plain_file.name, key, dest_filename)
        storage(settings.STORAGE_CACHE).import_filename(gz_plain_file.name, key, dest_filename + '.gz')



# coding=utf-8
__author__ = 'flanker'


class InvalidRepositoryException(BaseException):
    def __init__(self, msg):
        self.msg = msg

    def __unicode__(self):
        return self.msg

    def __str__(self):
        return self.msg



# coding=utf-8
import os
import re
from tarfile import InvalidHeaderError
import struct
import math
import io

__author__ = 'flanker'


class ArInfo(object):
    def __init__(self, name=''):
        self.name = name
        self.size = 0
        self.timestamp = 0
        self.owner_id = 0
        self.group_id = 0
        self.file_mode = 0


class ArObjFile(object):
    def __init__(self, fileobj, file_size, bufsize=10240):
        self.bufsize = bufsize
        self.fileobj = fileobj
        self.file_size = file_size
        self.seek_offset = self.fileobj.tell()
        self.read_size = 0

    @staticmethod
    def seekable():
        return True

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            self.read_size = offset
        elif whence == os.SEEK_CUR:
            self.read_size += offset
        elif whence == os.SEEK_END:
            self.read_size = self.file_size + offset
        self.fileobj.seek(self.read_size + self.seek_offset)

    def read(self, size=None):
        if size is None:
            size = self.file_size - self.read_size
        size = min(self.file_size - self.read_size, size)
        self.read_size += size
        return self.fileobj.read(size)

    def close(self):
        pass

    def tell(self):
        return self.read_size


class ArFile(object):
    GLOBAL_HEADER = b'!<arch>\n'

    def __init__(self, name=None, mode='r', fileobj=None, bufsize=10240):
        self.name = name
        self.mode = mode[0]
        self.fileobj = fileobj
        self.bufsize = bufsize
        if fileobj is None:
            self.fileobj = open(name, self.mode + 'b')
        if self.mode == 'r':
            if self.fileobj.read(8) != self.GLOBAL_HEADER:
                raise InvalidHeaderError
        else:
            self.fileobj.write(self.GLOBAL_HEADER)
        self.__current_pos = self.fileobj.tell()

    def __reset(self):
        self.fileobj.seek(0)
        if self.fileobj.read(8) != self.GLOBAL_HEADER:
            raise InvalidHeaderError
        self.__current_pos = 8

    def getmember(self, name):
        self.__reset()
        ar_info = self.next()
        while ar_info is not None and ar_info.name != name:
            ar_info = self.next()
        return ar_info

    def getmembers(self):
        self.__reset()
        members = []
        ar_info = self.next()
        while ar_info is not None:
            members.append(ar_info)
            ar_info = self.next()
        return members

    def getnames(self):
        self.__reset()
        members = []
        ar_info = self.next()
        while ar_info is not None:
            members.append(ar_info.name)

            ar_info = self.next()
        return members

    def next(self):
        self.fileobj.seek(self.__current_pos)
        data = self.fileobj.read(60)
        magic = data[58:60]  # byte
        if magic != b'\x60\x0a':
            return None
        assert isinstance(data, bytes)
        ar_info = ArInfo(data[0:16].decode('utf-8').strip())
        ar_info.filename = data[0:16].decode('utf-8').strip()
        ar_info.timestamp = data[16:28]
        ar_info.owner_id = data[28:34]
        ar_info.group_id = data[34:40]
        ar_info.file_mode = data[40:48]
        ar_info.size = int(data[48:58])
        self.__current_pos += (60 + int((ar_info.size + 1) / 2) * 2)
        return ar_info

    def extractall(self, path='.', members=None):
        members = set([x.name for x in members]) if members is not None else None
        self.__reset()
        member = self.next()
        while member is not None:
            if members is None or member.name in members:
                fullpath = os.path.join(path, member.name)
                e = ArObjFile(self.fileobj, member.size)
                with open(fullpath, 'wb') as fd:
                    data = e.read(self.bufsize)
                    while data:
                        fd.write(data)
                        data = e.read(self.bufsize)
            member = self.next()

    def extract(self, member, path=''):
        self.__reset()
        if not isinstance(member, ArInfo):
            member = ArInfo(name=member)
        return self.extractall(path=path, members=(member, ))

    def extractfile(self, member):
        name = member.name if isinstance(member, ArInfo) else member
        self.__reset()
        member = self.next()
        while member is not None:
            if member.name == name:
                return ArObjFile(self.fileobj, member.size)
            member = self.next()

    def add(self, name, arcname=None):
        pass

    def addfile(self, arinfo, fileobj=None):
        pass

    def getarinfo(self, name=None, arcname=None, fileobj=None):
        pass

    def close(self):
        pass



# coding=utf-8
from moneta import __version__ as version

__author__ = 'flanker'


def context_base(request):
    user_agent = request.META.get('HTTP_USER_AGENT', '').lower()
    is_os_x = 'macintosh' in user_agent
    is_linux = 'linux' in user_agent
    absolute_url = request.build_absolute_uri('/')[:-1]
    return {'absolute_url': absolute_url, 'use_https': absolute_url.startswith('https'),
            'is_linux': is_linux, 'is_os_x': is_os_x, 'moneta_version': version}



# -*- coding=utf-8 -*-
from django.conf import settings

__author__ = 'flanker'
import optparse

from django.core.management.base import BaseCommand, CommandError
from moneta.repository.signing import GPG


if settings.ADMINS and len(settings.ADMINS[0]) == 2:
    default_email = settings.ADMINS[0][1]
else:
    default_email = 'moneta@19pouces.net'


class Command(BaseCommand):
    args = '<generate|show|export>'
    help = """command=generate: Create a new GPG key
    command=show: Show existing GPG keys
    command=export: export GPG key"""
    option_list = BaseCommand.option_list + (
        optparse.make_option('--type', action='store', dest='type', default='RSA', help='Key type (RSA or DSA).'),
        optparse.make_option('--length', action='store', dest='length', default='1024',
                             help='Key length (default 1024).'),
        optparse.make_option('--name', action='store', dest='name', default='Moneta GNUPG key', help='Name of the key'),
        optparse.make_option('--comment', action='store', dest='comment', default='Generated by gnupg.py',
                             help='Comment to add to the generated key.'),
        optparse.make_option('--email', action='store', dest='email', default=default_email,
                             help='Email address for the user.'),
        optparse.make_option('--years', action='store', dest='years', default='10y',
                             help='Expiration date, in number of years (like "10y") or days (like "10d").'),
    )

    def handle(self, *args, **options):
        if len(args) == 0 or args[0] not in ('generate', 'show', 'export'):
            raise CommandError('Usage: gpg_gen <command>')
        command = args[0]
        if command == 'generate':
            input_data = GPG.gen_key_input(key_type=options['type'], key_length=int(options['length']),
                                           name_real=options['name'], name_comment=options['comment'],
                                           name_email=options['email'],
                                           expire_date=options['years'])
            key = GPG.gen_key(input_data)
            print("Fingerprint", key)
        elif command == 'show':
            print("Available keys:")
            for key in GPG.list_keys(False):
                print("id (GNUPG_KEYID) : {keyid}, longueur : {length}, empreinte : {fingerprint}".format(**key))
        elif command == 'export':
            print(GPG.export_keys(settings.GNUPG_KEYID))


"""
Utility module providing nice (colorized) logger.

"""
from bz2 import BZ2Decompressor
import gzip
import logging
import logging.handlers
import os
import shutil
import stat
import tempfile
import unicodedata
import zlib

from django.conf import settings
from django.core import exceptions
from django.http import QueryDict
from djangofloor.utils import import_module
from django.utils.translation import ugettext_lazy as _


def import_path(middleware_path):
    try:
        mw_module, mw_fnname = middleware_path.rsplit('.', 1)
    except ValueError:
        raise exceptions.ImproperlyConfigured('%s isn\'t a middleware module' % middleware_path)
    try:
        mod = import_module(mw_module)
    except ImportError as e:
        raise exceptions.ImproperlyConfigured('Error importing middleware %s: "%s"' % (mw_module, e))
    try:
        mw_fn = getattr(mod, mw_fnname)
    except AttributeError:
        raise exceptions.ImproperlyConfigured(
            'Middleware module "%s" does not define "%s"' % (mw_module, mw_fnname))
    return mw_fn


def remove(path):
    """
    Remove the path given in argument. All files must belong to the current user.
    :param path:
    """
    result = True
    if os.path.exists(path):
        # noinspection PyBroadException
        try:
            for root, dirs, files in os.walk(path, topdown=True):
                for name in files:
                    if not os.path.islink(os.path.join(root, name)):
                        os.chmod(os.path.join(root, name), stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                for name in dirs:
                    os.chmod(os.path.join(root, name), stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
            if os.path.isfile(path):
                os.remove(path)
            if os.path.isdir(path):
                shutil.rmtree(path)
        except Exception:
            result = False
            logging.warning(_('Unable to remove %(path)s.') % {'path': path, }, exc_info=True)
    return result


def rename(src, dst):
    result = True
    # noinspection PyBroadException
    try:
        if not makedir(dst):
            return False
        if not remove(dst):
            return False
        if os.path.isdir(src):
            shutil.copytree(src, dst)
        else:
            shutil.copy2(src, dst)
        remove(src)
    except Exception:
        result = False
        logging.warning(_('Unable to rename %(src)s to %(dst)s.') % {'src': src, 'dst': dst, }, exc_info=True)
    return result


def makedir(path):
    result = True
    # noinspection PyBroadException
    try:
        if not os.path.isdir(path):
            os.makedirs(path)
    except Exception:
        result = False
        logging.warning(_('Unable to create directory %(path)s.') % {'path': path, }, exc_info=True)
    return result


def normalize_str(x):
    """
    Normalize strings by changing uppercase chars to lowercase, and removing accents
    :param x: the string to normalize
    :return: the normalized string
    """
    return unicodedata.normalize('NFKD', x.lower()).encode('ASCII', 'ignore')


def mkdtemp():
    """
    create a temporary directory inside the settings.TEMP_ROOT
    :return: the name of the created directory
    """
    makedir(settings.TEMP_ROOT)
    return tempfile.mkdtemp(dir=settings.TEMP_ROOT, prefix='moneta')


def read_file_in_chunks(fileobj, chunk_size=4096):
    while True:
        data = fileobj.read(chunk_size)
        if not data:
            break
        yield data


class ZlibFile(object):
    def __init__(self, fileobj):
        self.__fileobj = fileobj
        self.__decompressor = zlib.decompressobj()
        self.__is_flushed = False

    def read(self, bufsize=10240):
        uncompressed_data = ''
        while not uncompressed_data and not self.__is_flushed:
            compressed_data = self.__fileobj.read(bufsize)
            if not compressed_data:
                self.__is_flushed = True
                uncompressed_data = self.__decompressor.flush()
            else:
                uncompressed_data = self.__decompressor.decompress(compressed_data)
        return uncompressed_data

    def close(self):
        pass


class BZ2File(object):
    def __init__(self, fileobj):
        self.__fileobj = fileobj
        self.__is_finished = False
        self.__decompressor = BZ2Decompressor()

    def read(self, bufsize):
        uncompressed_data = ''
        while not uncompressed_data and not self.__is_finished:
            compressed_data = self.__fileobj.read(bufsize)
            if not compressed_data:
                self.__is_finished = True
            else:
                uncompressed_data = self.__decompressor.decompress(compressed_data)
        return uncompressed_data

    def close(self):
        pass


def split_lines(fd, compression=None, bufsize=10240):
    __buffer = ''
    if compression == 'gz':
        tmp_file = tempfile.TemporaryFile(dir=settings.TEMP_ROOT)
        data = fd.read(bufsize)
        while data:
            tmp_file.write(data)
            data = fd.read(bufsize)
        tmp_file.seek(0)
        fd = gzip.GzipFile('filename', 'r', 9, fileobj=tmp_file)
    elif compression == 'bz2':
        fd = BZ2File(fd)
    data = fd.read(bufsize)
    while data:
        pos = data.find("\n")
        while pos > -1:
            if __buffer:
                yield __buffer + data[0:pos]
                __buffer = ''
            else:
                yield data[0:pos]
            data = data[pos + 1:]
            pos = data.find("\n")
        __buffer += data
        data = fd.read(bufsize)
    yield __buffer
    fd.close()


def parse_control_data(control_data, continue_line=' ', split=': ', skip_after_blank=False):
    offset = len(continue_line)
    result_data = QueryDict('', mutable=True)
    key, value = None, None
    description = ''
    add_to_description = False
    for line in control_data.splitlines():
        if not line.split() and skip_after_blank:
            add_to_description = True
        if add_to_description:
            description += "\n"
            description += line
            continue
        if not line or line[0:offset] == continue_line:
            if key is not None:
                value += "\n"
                value += line[offset:]
        else:
            if key is not None:
                if key not in result_data:
                    result_data.setlist(key, [value])
                else:
                    result_data.appendlist(key, value)
            key, value = line.split(split, 1)
            value = value.lstrip()
    if key is not None:
        if key not in result_data:
            result_data.setlist(key, [value])
        else:
            result_data.appendlist(key, value)
    if add_to_description:
        result_data.setlist('description', [description])
    return result_data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0006_require_contenttypes_0002'),
    ]

    operations = [
        migrations.CreateModel(
            name='ArchiveState',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=100, verbose_name='Nom')),
                ('slug', models.SlugField(max_length=100, verbose_name='Nom raccourci')),
                ('creation', models.DateTimeField(db_index=True, auto_now_add=True, verbose_name='Date de création')),
                ('modification', models.DateTimeField(db_index=True, auto_now=True, verbose_name='Date de modification')),
                ('author', models.ForeignKey(blank=True, null=True, verbose_name='Auteur', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Element',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=100, verbose_name='Nom')),
                ('slug', models.SlugField(max_length=100, verbose_name='Nom raccourci')),
                ('creation', models.DateTimeField(db_index=True, auto_now_add=True, verbose_name='Date de création')),
                ('modification', models.DateTimeField(db_index=True, auto_now=True, verbose_name='Date de modification')),
                ('official_link', models.URLField(max_length=255, blank=True, verbose_name='URL pour la page web')),
                ('short_description', models.CharField(max_length=500, blank=True, verbose_name='Description courte')),
                ('long_description', models.TextField(blank=True, verbose_name='Description longue')),
                ('full_name', models.CharField(db_index=True, max_length=255, blank=True, verbose_name='Nom complet')),
                ('full_name_normalized', models.CharField(db_index=True, max_length=255, blank=True, help_text='Nom complet, sans caractères spéciaux', verbose_name='Nom complet normalisé')),
                ('archive', models.CharField(db_index=True, max_length=255, blank=True, default='', verbose_name='Archive')),
                ('version', models.CharField(db_index=True, max_length=255, blank=True, default='', verbose_name='Version')),
                ('filename', models.CharField(max_length=255, blank=True, verbose_name='nom de fichier', db_index=True, default='', help_text='généré automatiquement à la création')),
                ('uuid', models.CharField(db_index=True, max_length=40, blank=True, help_text='Unique identifier, automatically generated on first save', verbose_name='UUID')),
                ('md5', models.CharField(max_length=120, blank=True, verbose_name='somme MD5', db_index=True, default='', help_text='généré automatiquement à la création')),
                ('sha1', models.CharField(max_length=120, blank=True, verbose_name='somme SHA1', db_index=True, default='', help_text='généré automatiquement à la création')),
                ('sha256', models.CharField(max_length=120, blank=True, verbose_name='somme SHA256', db_index=True, default='', help_text='généré automatiquement à la création')),
                ('filesize', models.IntegerField(db_index=True, blank=True, default=0, help_text='généré automatiquement à la création', verbose_name='taille')),
                ('extension', models.CharField(max_length=20, blank=True, verbose_name='extension', db_index=True, default='', help_text='généré automatiquement à la création')),
                ('mimetype', models.CharField(max_length=40, blank=True, verbose_name='MIME type', db_index=True, default='', help_text='Guessed on creation')),
                ('archive_file', models.FileField(blank=True, upload_to='uploads', null=True, verbose_name='Archive file', default='', help_text='Any text, binary or archive file.')),
                ('archive_key', models.CharField(db_index=True, max_length=255, blank=True, default='', verbose_name='Original file')),
                ('uncompressed_key', models.CharField(db_index=True, max_length=255, blank=True, default='', verbose_name='Stored path')),
                ('extra_data', models.TextField(blank=True, default='', verbose_name='Extra repo data')),
                ('author', models.ForeignKey(blank=True, null=True, verbose_name='Auteur', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'files',
                'verbose_name': 'file',
            },
        ),
        migrations.CreateModel(
            name='ElementSignature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('signature', models.TextField(blank=True, default='', verbose_name='signature')),
                ('method', models.CharField(db_index=True, max_length=10, choices=[('gpg', 'GnuPG'), ('openssl', 'OpenSSL/x509')], verbose_name='méthode de signature')),
                ('creation', models.DateTimeField(db_index=True, auto_now_add=True, verbose_name='Date de création')),
                ('element', models.ForeignKey(verbose_name='Élément', default=0, to='repository.Element')),
            ],
        ),
        migrations.CreateModel(
            name='Repository',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=100, verbose_name='Nom')),
                ('slug', models.SlugField(max_length=100, verbose_name='Nom raccourci')),
                ('creation', models.DateTimeField(db_index=True, auto_now_add=True, verbose_name='Date de création')),
                ('modification', models.DateTimeField(db_index=True, auto_now=True, verbose_name='Date de modification')),
                ('archive_type', models.CharField(db_index=True, max_length=100, verbose_name='Type de dépôt')),
                ('on_index', models.BooleanField(db_index=True, default=True, verbose_name='Display on public index?')),
                ('is_private', models.BooleanField(db_index=True, default=False, verbose_name='Authentification requise en lecture ?')),
                ('admin_group', models.ManyToManyField(db_index=True, related_name='repository_admin', blank=True, to='auth.Group', verbose_name="Groupes d'administration")),
                ('author', models.ForeignKey(blank=True, null=True, verbose_name='Auteur', to=settings.AUTH_USER_MODEL)),
                ('reader_group', models.ManyToManyField(db_index=True, related_name='repository_reader', blank=True, to='auth.Group', verbose_name='Groupes ayant accès en lecture')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='element',
            name='repository',
            field=models.ForeignKey(to='repository.Repository', verbose_name='Dépôt'),
        ),
        migrations.AddField(
            model_name='element',
            name='states',
            field=models.ManyToManyField(db_index=True, to='repository.ArchiveState', verbose_name='États possibles'),
        ),
        migrations.AddField(
            model_name='archivestate',
            name='repository',
            field=models.ForeignKey(to='repository.Repository', verbose_name='Dépôt'),
        ),
    ]



# -*- coding: utf-8 -*-
__author__ = 'flanker'
from djangofloor.iniconf import OptionParser, bool_setting


def x_accel_converter(value):
    if bool_setting(value):
        return [('{LOCAL_PATH}/storage/', '{MEDIA_URL}'), ]
    return []


INI_MAPPING = [
    OptionParser('SERVER_NAME', 'global.server_name'),
    OptionParser('PROTOCOL', 'global.protocol'),
    OptionParser('BIND_ADDRESS', 'global.bind_address'),
    OptionParser('LOCAL_PATH', 'global.data_path'),
    OptionParser('ADMIN_EMAIL', 'global.admin_email'),
    OptionParser('TIME_ZONE', 'global.time_zone'),
    OptionParser('LANGUAGE_CODE', 'global.language_code'),
    OptionParser('USE_X_SEND_FILE', 'global.x_send_file', bool_setting),
    OptionParser('X_ACCEL_REDIRECT', 'global.x_accel_converter', x_accel_converter),
    OptionParser('FLOOR_AUTHENTICATION_HEADER', 'global.remote_user_header'),
    OptionParser('EXTRA_INSTALLED_APP', 'global.extra_app'),

    OptionParser('GNUPG_HOME', 'gnupg.home'),
    OptionParser('GNUPG_KEYID', 'gnupg.keyid'),
    OptionParser('GNUPG_PATH', 'gnupg.path'),

    OptionParser('DATABASE_ENGINE', 'database.engine'),
    OptionParser('DATABASE_NAME', 'database.name'),
    OptionParser('DATABASE_USER', 'database.user'),
    OptionParser('DATABASE_PASSWORD', 'database.password'),
    OptionParser('DATABASE_HOST', 'database.host'),
    OptionParser('DATABASE_PORT', 'database.port'),

]



# coding=utf-8
from django.contrib import messages
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _


def public_check(request):
    messages.success(request, _('You can access to this page.'))
    return render_to_response('core/empty.html', RequestContext(request))


def private_check(request):
    if request.user.is_anonymous():
        messages.error(request, _('You are not authenticated.'))
    else:
        messages.success(request, _('You can access to this page and you are authenticated.'))
    return render_to_response('core/empty.html', RequestContext(request))



# -*- coding: utf-8 -*-
import json
import tempfile
import time

from django.conf import settings
from django.conf.urls import url
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
# noinspection PyPackageRequirements
from django.utils.translation import ugettext as _

from moneta.repositories.aptitude import Aptitude
from moneta.repositories import rpm
from moneta.repository.models import Repository, storage, Element, ArchiveState
from moneta.repository.signing import GPGSigner
from moneta.views import sendpath

__author__ = 'flanker'


class Yum(Aptitude):
    verbose_name = _('YUM repository for Linux .rpm packages')
    storage_uid = 'a87172de-0000-0000-0000-%012d'
    archive_type = 'yum'
    index_html = 'repositories/yum/index.html'

    def is_file_valid(self, uploaded_file):
        if not uploaded_file.name.endswith('.rpm'):
            return False
        try:
            rpm.RPM(uploaded_file.file)
        except rpm.RPMError:
            return False
        return True

    def update_element(self, element):
        fd = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key, sub_path='')
        rpm_obj = rpm.RPM(fd)
        element.filename = rpm_obj.canonical_filename
        element.version = rpm_obj.header.version
        element.archive = rpm_obj.header.name
        header = {}
        signature = {}
        for (obj_dict, header_base) in ((header, rpm_obj.header), (signature, rpm_obj.signature)):
            available = {}
            for entry in header_base:
                available[entry.tag] = entry.value
            for attr_name, infos in header_base.TAGS.items():
                attr_value = available.get(infos[0], infos[1])
                if not isinstance(attr_value, bytes):
                    obj_dict[attr_name] = attr_value
        rpm_ = {'binary': rpm_obj.binary, 'canonical_filename': rpm_obj.canonical_filename,
                'checksum': rpm_obj.checksum, 'filesize': rpm_obj.filesize, 'source': rpm_obj.source,
                'filelist': [{'type': x.type, 'name': x.name, } for x in rpm_obj.filelist],
                'provides': [{'name': x.name, 'str_flags': x.str_flags, 'version': list(x.version)} for x in rpm_obj.provides],
                'requires': [{'name': x.name, 'str_flags': x.str_flags, 'version': list(x.version)} for x in rpm_obj.requires],
                'changelog': [{'name': x.name, 'time': x.time, 'text': x.text, } for x in rpm_obj.changelog],
                'obsoletes': [{'name': x.name, 'str_flags': x.str_flags, 'version': list(x.version)} for x in rpm_obj.obsoletes],
                'conflicts': [{'name': x.name, 'str_flags': x.str_flags, 'version': list(x.version)} for x in rpm_obj.conflicts],
                'header_range': list(rpm_obj.header.header_range),
                }
        rpm_dict = {'header': header, 'signature': signature, 'rpm': rpm_, }
        element.extra_data = json.dumps(rpm_dict)

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            * /1/pool/repo_apt/stable/[package-1.0.0-amd64.deb]
            * /1/dists/repo_apt/Release[.gpg]
            * /1/dists/repo_apt/Contents-amd64.gz
            * /1/dists/repo_apt/stable/Release.xz
            * /1/dists/repo_apt/stable/binary-amd64/Packages.gz
            * /1/dists/repo_apt/stable/binary-amd64/Release.bz2
            * /1/keys/repo_apt/key.asc

        :return: a patterns as expected by django

        """
        pattern_list = [
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/(?P<folder>[\w\-\._]+)/Packages/'
                r'(?P<filename>[\w\-\.]+)$', self.wrap_view('get_file'), name='get_file'),
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/(?P<arch>[\w\-\._]+)/repodata/'
                r'(?P<filename>\w+\.xml)(?P<compression>|.bz2|.gz)$', self.wrap_view('repodata_file'), name='repodata_file'),
            url(r'^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/(?P<arch>[\w\-\._]+)$',
                self.wrap_view('index'), name='repo_index'),
            url(r"^(?P<rid>\d+)/gpg_key.asc$", self.wrap_view('gpg_key'), name="gpg_key"),
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def gpg_key(self, request, rid, repo_slug=None, slug=None):
        get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        signature = GPGSigner().export_key()
        # noinspection PyUnusedLocal
        repo_slug, slug = slug, repo_slug
        return HttpResponse(signature, content_type="text/plain")

    def repodata_file(self, request, rid, repo_slug, state_slug, arch, filename, compression):
        if filename not in ('comps.xml', 'primary.xml', 'other.xml', 'filelists.xml', 'repomd.xml', ):
            return HttpResponse(_('File not found'), status=404)
        if compression and filename == 'repomd.xml':
            return HttpResponse(_('File not found'), status=404)
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        filename = self.index_filename(state_slug, arch, filename + compression)
        mimetype = 'text/xml'
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        uid = self.storage_uid % repo.id
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        return sendpath(settings.STORAGE_CACHE, key, filename, mimetype)

    def generate_indexes(self, repository, states=None, validity=365):
        if states is None:
            states = list(ArchiveState.objects.filter(repository=repository).order_by('name'))
        revision = int(time.time())
        architectures_by_state = {x.slug: set() for x in states}  # architectures_by_state[archive_state.slug] = {'x86_64', 'c7', }
        # load all dict infos and count all architectures
        rpm_objects = []
        package_count_by_state_arch = {x.slug: {'noarch': 0} for x in states}
        for element in Element.objects.filter(repository=repository).prefetch_related('states'):
            rpm_dict = json.loads(element.extra_data)
            rpm_objects.append(rpm_dict)
            rpm_dict['states'] = [s.slug for s in element.states.all()]
            package_architecture = rpm_dict['header']['architecture'] or 'noarch'
            if package_architecture != 'noarch':
                for state_slug in rpm_dict['states']:
                    architectures_by_state[state_slug].add(package_architecture)
            for state_slug in rpm_dict['states']:
                package_count_by_state_arch[state_slug].setdefault(package_architecture, 0)
                package_count_by_state_arch[state_slug][package_architecture] += 1

        # add the count of 'noarch' packages to other architectures
        for state_slug, package_count_by_arch in package_count_by_state_arch.items():
            if len(package_count_by_arch) == 1:  # only 'noarch' architecture
                package_count_by_arch['x86_64'] = 0
                architectures_by_state[state_slug] = {'x86_64', }
            noarch_count = package_count_by_arch['noarch']
            del package_count_by_arch['noarch']
            for architecture in package_count_by_arch:
                package_count_by_arch[architecture] += noarch_count

        # prepare all files
        open_files = {}
        for state_slug, architectures in architectures_by_state.items():
            for architecture in architectures:
                def write(name_, data):
                    filename_ = self.index_filename(state_slug, architecture, name_)
                    open_files[filename_].write(data.encode('utf-8'))
                for name in ('other.xml', 'filelists.xml', 'comps.xml', 'primary.xml', ):

                    filename = self.index_filename(state_slug, architecture, name)
                    open_files[filename] = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
                    write(name, '<?xml version="1.0" encoding="UTF-8"?>\n')
                package_count = package_count_by_state_arch[state_slug][architecture]
                write('other.xml', '<otherdata xmlns="http://linux.duke.edu/metadata/other" packages="%d">\n' % package_count)
                write('filelists.xml', '<filelists xmlns="http://linux.duke.edu/metadata/filelists" packages="%d">\n' % package_count)
                write('comps.xml', '<!DOCTYPE comps PUBLIC "-//CentOS//DTD Comps info//EN" "comps.dtd">\n')
                write('comps.xml', '<comps>\n')
                write('primary.xml', '<metadata xmlns="http://linux.duke.edu/metadata/common" xmlns:rpm="http://linux.duke.edu/metadata/rpm" packages="%d">\n' % package_count)
        # fill all files with RPMs
        for rpm_dict in rpm_objects:
            filelists = render_to_string('repositories/yum/filelists.xml', rpm_dict)
            primary = render_to_string('repositories/yum/primary.xml', rpm_dict)
            other = render_to_string('repositories/yum/other.xml', rpm_dict)
            for state_slug in rpm_dict['states']:
                architectures = {rpm_dict['header']['architecture'], }
                if architectures == {'noarch', }:
                    architectures = architectures_by_state[state_slug]
                for architecture in architectures:
                    open_files[self.index_filename(state_slug, architecture, 'filelists.xml')].write(filelists.encode('utf-8'))
                    open_files[self.index_filename(state_slug, architecture, 'primary.xml')].write(primary.encode('utf-8'))
                    open_files[self.index_filename(state_slug, architecture, 'other.xml')].write(other.encode('utf-8'))
        # finish all files
        for state_slug, architectures in architectures_by_state.items():
            for architecture in architectures:
                open_files[self.index_filename(state_slug, architecture, 'other.xml')].write(b'</otherdata>')
                open_files[self.index_filename(state_slug, architecture, 'filelists.xml')].write(b'</filelists>')
                open_files[self.index_filename(state_slug, architecture, 'comps.xml')].write(b'</comps>')
                open_files[self.index_filename(state_slug, architecture, 'primary.xml')].write(b'</metadata>')

        storage_uid = self.storage_uid % repository.id
        # generate a compressed version of each file
        list_of_hashes = self.compress_files(open_files, '', storage_uid)
        dict_of_hashes = {x[0]: x for x in list_of_hashes}
        for state_slug, architectures in architectures_by_state.items():
            for architecture in architectures:
                filename = self.index_filename(state_slug, architecture, 'repomd.xml')
                open_files[filename] = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
                other = self.index_filename(state_slug, architecture, 'other.xml')
                filelists = self.index_filename(state_slug, architecture, 'filelists.xml')
                comps = self.index_filename(state_slug, architecture, 'comps.xml')
                primary = self.index_filename(state_slug, architecture, 'primary.xml')
                template_values = {'revision': revision,
                                   'other': dict_of_hashes[other],
                                   'filelists': dict_of_hashes[filelists],
                                   'comps': dict_of_hashes[comps],
                                   'primary': dict_of_hashes[primary],
                                   'other_gz': dict_of_hashes[other + '.gz'],
                                   'filelists_gz': dict_of_hashes[filelists + '.gz'],
                                   'comps_gz': dict_of_hashes[comps + '.gz'],
                                   'primary_gz': dict_of_hashes[primary + '.gz'], }
                repomd = render_to_string('repositories/yum/repomd.xml', template_values)
                repomd_file = open_files[filename]
                repomd_file.write(repomd.encode('utf-8'))
                repomd_file.flush()
                repomd_file.seek(0)
                storage(settings.STORAGE_CACHE).store_descriptor(storage_uid, filename, repomd_file)

    @staticmethod
    def index_filename(state: str, architecture: str, name: str):
        return '%(state)s/%(architecture)s/repodata/%(name)s' % {'state': state, 'architecture': architecture, 'name': name, }



# coding=utf-8
from django.core.management import BaseCommand
from moneta.repository.models import Element

__author__ = 'flanker'


class Command(BaseCommand):
    def handle(self, *args, **options):
        for elt in Element.objects.all():
            elt.delete()


# -*- coding: utf-8 -*-
import subprocess
from django.test import TestCase
import io
import pkg_resources
import yaml
from moneta.repositories.ruby import RubyLoader, RubyGem
from moneta.repositories.tests import RepositoryTestCase

__author__ = 'Matthieu Gallet'

metadata = """--- !ruby/object:Gem::Specification
name: m
version: !ruby/object:Gem::Version
  version: 1.4.0
platform: ruby
authors:
- Nick Quaranto
autorequire:
bindir: bin
cert_chain: []
date: 2015-09-27 00:00:00.000000000 Z
dependencies:
- !ruby/object:Gem::Dependency
  name: method_source
  requirement: !ruby/object:Gem::Requirement
    requirements:
    - - ">="
      - !ruby/object:Gem::Version
        version: 0.6.7
  type: :runtime
  prerelease: false
  version_requirements: !ruby/object:Gem::Requirement
    requirements:
    - - ">="
      - !ruby/object:Gem::Version
        version: 0.6.7
- !ruby/object:Gem::Dependency
  name: appraisal
  requirement: !ruby/object:Gem::Requirement
    requirements:
    - - ">="
      - !ruby/object:Gem::Version
        version: '0'
  type: :development
  prerelease: false
  version_requirements: !ruby/object:Gem::Requirement
    requirements:
    - - ">="
      - !ruby/object:Gem::Version
        version: '0'
description: Run test/unit tests by line number. Metal!
email:
- nick@quaran.to
executables:
- m
extensions: []
extra_rdoc_files: []
files:
- Appraisals
- Gemfile
- Gemfile.lock
- LICENSE
homepage: https://github.com/qrush/m
licenses: []
metadata: {}
post_install_message:
rdoc_options: []
require_paths:
- lib
required_ruby_version: !ruby/object:Gem::Requirement
  requirements:
  - - ">="
    - !ruby/object:Gem::Version
      version: '1.9'
required_rubygems_version: !ruby/object:Gem::Requirement
  requirements:
  - - ">="
    - !ruby/object:Gem::Version
      version: '0'
requirements: []
rubyforge_project:
rubygems_version: 2.4.8
signing_key:
specification_version: 4
summary: Run test/unit tests by line number. Metal!
test_files:
- test/Rakefile
- test/active_support_test.rb
"""


class TestMetadata(TestCase):

    def test_metadata(self):
        stream = io.BytesIO(metadata.encode('utf-8'))
        values = yaml.load(stream, Loader=RubyLoader)
        self.assertEqual('2.4.8', values.values['rubygems_version'])


class TestRubyScript(TestCase):

    def test_script(self):
        filename = pkg_resources.resource_filename('moneta.repositories', 'ruby.rb')
        # p = subprocess.Popen(['ruby'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        # p.communicate()


class TestRubyGem(RepositoryTestCase):

    def test_add_file(self):
        repo = self.create_repository(RubyGem)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'm-1.4.0.gem')
        self.add_file_to_repository(repo, filename)

    def test_add_file_2(self):
        repo = self.create_repository(RubyGem)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'method_source-0.8.2.gem')
        self.add_file_to_repository(repo, filename)

    def test_generate_index(self):
        repo = self.create_repository(RubyGem)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'm-1.4.0.gem')
        self.add_file_to_repository(repo, filename)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'method_source-0.8.2.gem')
        self.add_file_to_repository(repo, filename)
        ruby = RubyGem()
        ruby.generate_indexes(repo)



# -*- coding: utf-8 -*-
__author__ = 'flanker'
# coding=utf-8

from django.conf.urls import include, url
from django.contrib import admin

admin.autodiscover()

urls = [
    url(r'^repo/', include('moneta.repository.urls')),
    url(r'^core/', include('moneta.urls', namespace='moneta')),
]



# coding=utf-8
import mimetypes
import os
import shutil
from moneta.utils import makedir, remove

__author__ = 'flanker'

mimetypes.init()


class BaseStorage(object):
    # noinspection PyPep8Naming
    def __init__(self, ENGINE=''):
        """
        Initialise the storage backend
        :param ENGINE: class of this storage backend
        """
        self.engine = ENGINE

    def store(self, uid, path):
        """
        Store the directory in its internal storage (can be a zip file, or flat files)
        :param uid: UUID of the Element
        :param path: absolute path to store
        :return: a key unique to this storage
        :raise:
        """
        raise NotImplementedError

    def list_dir(self, key, sub_path):
        """
        List elements (directories and files) in a given directory
        :param key: UUID of the Element
        :param sub_path: relative path of the directory to list
        :return: A tuple of lists (directory names, file names)
        :raise:
        """
        for root, dirs, files in self.walk(key, sub_path):
            return dirs, files

    def walk(self, key, sub_path):
        """
        List elements (directories and files) in a given directory
        :param key: UUID of the Element
        :param sub_path: relative path of the directory to list
        :return: Same thing as os.walk
        :raise:
        """
        raise NotImplementedError

    def store_descriptor(self, uid, filename, fd):
        """
        Store a file content given by a file descriptor
        :param uid: UUID of the Element
        :param fd: file descriptor
        :return: a key unique to this storage
        :raise:
        """
        raise NotImplementedError

    def get_file(self, key, sub_path='', mode='rb'):
        """
        Return a file descriptor in read mode of the given path
        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: A file-like object
        :raise:
        """
        raise NotImplementedError

    def get_path(self, key, sub_path):
        """ return an absolute path of the file, or None if does not exist (e.g., database storage).

        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: `str`
        :raise:
        """
        raise NotImplementedError

    def get_relative_path(self, key, sub_path):
        """ return an relative path of the file to the root, or None if does not exist (e.g., database storage).

        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: `str`
        :raise:
        """
        raise NotImplementedError

    def delete(self, key):
        """
        Delete the element from its internal storage
        :param key: UUID of the Element to delete
        :raise:
        """
        raise NotImplementedError

    def import_filename(self, filename, key, sub_path=''):
        """import the filename to the storage. Remove the local source filename!

        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return:
        :rtype: :class:`bool`
        """
        raise NotImplementedError

    def get_size(self, key, sub_path):
        """
        Return the file size
        :param key: UUID of the Element to delete
        :raise:
        """
        raise NotImplementedError

    def uid_to_key(self, uid):
        raise NotImplementedError

    def mimetype(self, key, sub_path):
        raise NotImplementedError


class FlatStorage(BaseStorage):
    # noinspection PyPep8Naming
    def __init__(self, ENGINE='', ROOT=None, PATH_LEN=1):
        self.root = os.path.abspath(ROOT)
        self.path_len = PATH_LEN
        super(FlatStorage, self).__init__(ENGINE=ENGINE)

    def uid_to_key(self, uid):
        return os.path.join(*(self.split_uid(uid) + [uid]))

    def store(self, uid, path):
        """
        Store the directory in its internal storage (can be a zip file, or flat files)
        :param uid: UUID of the Element
        :param path: absolute path to store
        :return: a key unique to this storage
        :raise:
        """
        components = [self.root] + self.split_uid(uid) + [uid]
        components.append(os.path.basename(path))
        shutil.copytree(path, os.path.join(*components))
        return self.uid_to_key(uid)

    def list_dir(self, key, sub_path):
        """
        List elements (directories and files) in a given directory
        :param key: UUID of the Element
        :param sub_path: relative path of the directory to list
        :return: A tuple of lists (directory names, file names)
        :raise:
        """
        for root, dirs, files in self.walk(key, sub_path):
            return dirs, files

    def walk(self, key, sub_path):
        """
        List elements (directories and files) in a given directory
        :param key: UUID of the Element
        :param sub_path: relative path of the directory to list
        :return: Same behaviour as os.walk
        :raise:
        """
        components = [self.root, key]
        root_path = os.path.abspath(os.path.join(*components))
        if sub_path:
            components.append(sub_path)
        path = os.path.abspath(os.path.join(*components))
        if os.path.isfile(path):
            return self.simple_generator(
                (os.path.relpath(os.path.dirname(path), root_path), (), (os.path.basename(path), )))
        else:
            return self.rel_generator(os.walk(path), root_path)

    # noinspection PyMethodMayBeStatic
    def rel_generator(self, walker, root_path):
        for root, dirs, files in walker:
            yield os.path.relpath(root, root_path), dirs, files

    # noinspection PyMethodMayBeStatic
    def simple_generator(self, lis):
        yield lis

    def get_file(self, key, sub_path='', mode='rb'):
        """
        Return a file descriptor in read mode of the given path
        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: A file-like object
        :raise:
        """
        components = [self.root, key]
        if sub_path:
            components.append(sub_path)
        abs_path = os.path.join(*components)
        makedir(os.path.dirname(abs_path))
        try:
            fd = open(abs_path, mode)
        except IOError:
            fd = None
        return fd

    def import_filename(self, filename, key, sub_path=''):
        components = [self.root, key]
        if sub_path:
            components.append(sub_path)
        abs_path = os.path.join(*components)
        makedir(os.path.dirname(abs_path))
        return os.rename(filename, abs_path)

    def get_path(self, key, sub_path):
        """ return an absolute path of the file, or None if does not exist (e.g., database storage).

        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: `str`
        :raise:
        """
        components = [self.root, key]
        if sub_path:
            components.append(sub_path)
        return os.path.join(*components)

    def get_relative_path(self, key, sub_path):
        """ return an relative path of the file to the root, or None if does not exist (e.g., database storage).

        :param key: UUID of the Element
        :param sub_path: relative path to read
        :return: `str`
        :raise:
        """
        return os.path.relpath(self.get_path(key, sub_path), self.root)

    def get_size(self, key, sub_path):
        components = [self.root, key]
        if sub_path:
            components.append(sub_path)
        return os.path.getsize(os.path.join(*components))

    def delete(self, key):
        """
        Delete the element from its internal storage
        :param key: UUID of the Element to delete
        :raise:
        """
        abs_path = os.path.join(self.root, key)

        def remove_dir(dirname):
            """
            recursively delete empty directories
            :param dirname: directory to remove
            :return: None
            """
            if not os.path.exists(dirname):
                return
            if len(os.listdir(dirname)) > 0 or dirname == self.root:
                return
            remove(dirname)
            remove_dir(os.path.dirname(dirname))

        result = remove(abs_path)
        remove_dir(os.path.dirname(abs_path))
        return result

    def store_descriptor(self, uid, filename, fd):
        """
        Store a file content given by a file descriptor
        :param uid: UUID of the Element
        :param fd: file descriptor
        :return: a key unique to this storage
        :raise:
        """
        components = [self.root] + self.split_uid(uid) + [uid, filename]
        abs_path = os.path.join(*components)
        makedir(os.path.dirname(abs_path))
        with open(abs_path, 'wb') as fd_write:
            data = fd.read(10240)
            while data:
                fd_write.write(data)
                data = fd.read(10240)
        return os.path.join(*(components[1:]))

    def split_uid(self, uid):
        return [x for x in uid[0:self.path_len]]

    def mimetype(self, key, sub_path):
        path = self.get_path(key, sub_path)
        return mimetypes.guess_type(path)[0] or 'application/octet-stream'


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
PyRPM
=====

PyRPM is a pure python, simple to use, module to read information from a RPM file.
"""

from collections import namedtuple
import hashlib
import re
import stat
import struct
import sys

if sys.version < '3':
    try:
        from io import StringIO as BytesIO
    except ImportError:
        from io import StringIO as BytesIO
else:
    from io import BytesIO


class Entry(object):
    """ RPM Header Entry """

    # noinspection PyShadowingBuiltins
    def __init__(self, entry=None, store=None, tag=None, type=None, value=None):
        # noinspection PyPep8Naming
        DECODING_MAP = {
            0: self._read_null,
            1: self._read_char,
            2: self._read_int8,
            3: self._read_int16,
            4: self._read_int32,
            5: self._read_int64,
            6: self._read_string,
            7: self._read_binary,
            8: self._read_string_array,
            9: self._read_string,
        }

        # read from file if possible
        if entry is not None and store is not None:
            # seek to position in store
            store.seek(entry[2])

            # decode information
            self.entry = entry
            self.tag = entry[0]
            self.type = entry[1]
            self.value = DECODING_MAP[entry[1]](store, entry[3])
        else:
            self.tag = tag
            self.type = type
            self.value = value

    def __str__(self):
        return "(%s, %s)" % (self.tag, self.value, )

    def __repr__(self):
        return "(%s, %s)" % (self.tag, self.value, )

    @staticmethod
    def _read_format(fmt, store):
        size = struct.calcsize(fmt)
        data = store.read(size)
        unpacked_data = struct.unpack(fmt, data)
        return unpacked_data[0] if len(unpacked_data) == 1 else unpacked_data

    # noinspection PyUnusedLocal
    @staticmethod
    def _read_null(store, data_count):
        return None

    def _read_char(self, store, data_count):
        """ store is a pointer to the store offset
        where the char should be read
        """
        return self._read_format('!{0:d}s'.format(data_count), store)

    def _read_int8(self, store, data_count):
        """ int8 = 1byte
        """
        return int(self._read_char(store, data_count))

    def _read_int16(self, store, data_count):
        """ int16 = 2bytes
        """
        return self._read_format('!{0:d}h'.format(data_count), store)

    def _read_int32(self, store, data_count):
        """ int32 = 4bytes
        """
        return self._read_format('!{0:d}i'.format(data_count), store)

    def _read_int64(self, store, data_count):
        """ int64 = 8bytes
        """
        return self._read_format('!{0:d}q'.format(data_count), store)

    def _read_string(self, store, data_count):
        """ read a string entry
        """
        string = b''
        while True:
            char = self._read_char(store, 1)
            if char == b'\x00':  # read until '\0'
                break
            string += char
        return string.decode('utf-8')

    def _read_string_array(self, store, data_count):
        """ read a array of string entries
        """
        return [self._read_string(store, 1) for i in range(data_count)]

    def _read_binary(self, store, data_count):
        """ read a binary entry
        """
        return self._read_format('!{0:d}s'.format(data_count), store)


# noinspection PyBroadException
class HeaderBase(object):

    """ RPM Header Structure """
    MAGIC_NUMBER = b'\x8e\xad\xe8'
    MAGIC_NUMBER_MATCHER = re.compile(b'(\x8e\xad\xe8)')

    TAGS = {}

    def __init__(self, file):
        """ read a RPM header structure with all its entries

            Header format:
            [3bytes][1byte][4bytes][4bytes][4bytes]
              MN      VER   UNUSED  IDXNUM  STSIZE

            Entry format:
            [4bytes][4bytes][4bytes][4bytes]
               TAG    TYPE   OFFSET  COUNT
        """
        self.entries = []

        # read from file if possible
        if file:
            # read and check header
            start = file.tell()
            header = struct.unpack('!3sc4sll', file.read(16))
            if header[0] != self.MAGIC_NUMBER:
                raise RPMError('invalid RPM header')

            # read entries and store
            entries = [file.read(16) for i in range(header[3])]
            store = BytesIO(file.read(header[4]))

            # parse entries
            for entry in entries:
                parsed_entry = struct.unpack("!4l", entry)
                object_entry = Entry(parsed_entry, store)

                if object_entry:
                    self.entries.append(object_entry)
            end = file.tell()
            self.header_range = (start, end)

    def __getattr__(self, name):
        if name in self.TAGS:
            id_, default = self.TAGS[name]
            try:
                return self[id_]
            except:
                return default

        raise AttributeError(name)

    def __iter__(self):
        for entry in self.entries:
            yield entry

    def __getitem__(self, item):
        for entry in self:
            if entry.tag == item:
                return entry.value
        raise KeyError()


# signature header section
class Signature(HeaderBase):
    TAGS = {
        'size': (1000, -1),
        'pgp': (1002, ""),
        'md5': (1004, ""),
        'gpg': (1005, ""),
        'pgp5': (1006, ""),
        'payload_size': (1007, -1),
    }


# primary header section
class Header(HeaderBase):
    TAGS = {
        'name': (1000, ""),
        'version': (1001, "0.1"),
        'release': (1002, ""),
        'epoch': (1003, 0),
        'summary': (1004, ""),
        'description': (1005, ""),
        'build_time': (1006, 0),
        'build_host': (1007, ""),
        'size': (1009, 0),
        'vendor': (1011, ""),
        'license': (1014, ""),
        'packager': (1015, ""),
        'group': (1016, []),
        'url': (1020, ""),
        'architecture': (1022, ""),
        'source_rpm': (1044, ""),
        'archive_size': (1046, 0),
        'provides': (1047, []),
        'requires': (1049, []),
        'conflicts': (1054, []),
        'platform': (1132, ""),
    }


class RPMError(BaseException):
    pass


RPMFile = namedtuple("RPMFile", ['name', 'size', 'mode', 'rdevice', 'device', 'time', 'digest', 'link_to',
                     'flags', 'username', 'group', 'verify_flags', 'language', 'inode', 'color', 'content_class', 'type', 'primary'])
RPMChangeLog = namedtuple("RPMChangeLog", ['name', 'text', 'time'])
RPMprco = namedtuple("RPMprco", ['name', 'version', 'flags', 'str_flags'])


# noinspection PyPep8Naming,PyBroadException
class RPM(object):
    RPM_LEAD_MAGIC_NUMBER = b'\xed\xab\xee\xdb'
    RPM_PRCO_FLAGS_MAP = {0: None, 2: 'LT', 4: 'GT', 8: 'EQ', 10: 'LE', 12: 'GE'}

    def __init__(self, rpm):
        """ rpm - StringIO.StringIO/io.BytesIO | file
        """
        if hasattr(rpm, 'read'):  # if it walk like a duck..
            self.rpmfile = rpm
        else:
            raise ValueError('invalid initialization: StringIO/BytesIO or file expected received %s' % (type(rpm), ))

        self.binary = None
        self.source = None
        self.header = None
        self.signature = None
        self.filelist = []
        self.changelog = []

        self.provides = []
        self.requires = []
        self.obsoletes = []
        self.conflicts = []

        self._read_lead()
        self._read_signature()
        self._read_header()
        self._match_composite()
        self._compute_checksum()

    @property
    def canonical_filename(self):
        if self.header.epoch == 0:
            return "%s-%s-%s.%s.rpm" % (self.header.name, self.header.version, self.header.release, self.header.architecture if self.binary else "src")
        else:
            return "%s-%s-%s-%d.%s.rpm" % (self.header.name, self.header.version, self.header.release, self.header.epoch, self.header.architecture if self.binary else "src")

    def _read_lead(self):
        """ reads the rpm lead section

            struct rpmlead {
               unsigned char magic[4];
               unsigned char major, minor;
               short type;
               short archnum;
               char name[66];
               short osnum;
               short signature_type;
               char reserved[16];
               } ;
        """
        lead_fmt = '!4sBBhh66shh16s'
        data = self.rpmfile.read(96)
        value = struct.unpack(lead_fmt, data)

        magic_num = value[0]
        ptype = value[3]

        if magic_num != self.RPM_LEAD_MAGIC_NUMBER:
            raise RPMError('wrong magic number this is not a RPM file')

        if ptype == 1:
            self.binary = False
            self.source = True
        elif ptype == 0:
            self.binary = True
            self.source = False
        else:
            raise RPMError('wrong package type this is not a RPM file')

    def _read_signature(self):
        """ read signature header """

        # find the start of the header
        if not self._find_magic_number():
            raise RPMError('invalid RPM file, signature area not found')

        # consume signature area
        self.signature = Signature(self.rpmfile)

    def _read_header(self):
        """ read information header """

        # find the start of the header
        if not self._find_magic_number():
            raise RPMError('invalid RPM file, header not found')

        # consume header area
        self.header = Header(self.rpmfile)

    def _find_magic_number(self):
        """ find a magic number in a buffer
        """
        string = self.rpmfile.read(1)
        while True:
            match = HeaderBase.MAGIC_NUMBER_MATCHER.search(string)
            if match:
                self.rpmfile.seek(-3, 1)
                return True
            byte = self.rpmfile.read(1)
            if not byte:
                return False
            else:
                string += byte
        return False

    def _match_composite(self):
        # files
        try:
            for idx, name in enumerate(self.header[1117]):
                dirname = self.header[1118][self.header[1116][idx]]
                self.filelist.append(RPMFile(
                    name=dirname + name,
                    size=self.header[1028][idx],
                    mode=self.header[1030][idx],
                    rdevice=self.header[1033][idx],
                    time=self.header[1034][idx],
                    digest=self.header[1035][idx],
                    link_to=self.header[1036][idx],
                    flags=self.header[1037][idx],
                    username=self.header[1039][idx],
                    group=self.header[1040][idx],
                    verify_flags=self.header[1045][idx],
                    device=self.header[1095][idx],
                    inode=self.header[1096][idx],
                    language=self.header[1097][idx],
                    color=self.header[1140][idx] if 1140 in self.header else None,
                    content_class=self.header[1142][self.header[1141][idx]] if 1142 in self.header and 1141 in self.header else None,
                    type='dir' if stat.S_ISDIR(self.header[1030][idx] & 65535) else ('ghost' if (self.header[1037][idx] & 64) else 'file'),
                    primary=('bin/' in dirname or dirname.startswith('/etc/'))))
        except:
            pass

        # change log
        try:
            if self.header[1081]:
                for name, time, text in zip(self.header[1081], self.header[1080], self.header[1082]):
                    self.changelog.append(RPMChangeLog(name=name, time=time, text=text))
        except:
            pass

        # provides
        try:
            if self.header[1047]:
                for name, flags, version in zip(self.header[1047], self.header[1112], self.header[1113]):
                    self.provides.append(
                        RPMprco(name=name, flags=flags, str_flags=self.RPM_PRCO_FLAGS_MAP[flags & 0xf], version=self._stringToVersion(version)))
        except:
            pass

        # requires
        try:
            if self.header[1049]:
                for name, flags, version in zip(self.header[1049], self.header[1048], self.header[1050]):
                    self.requires.append(
                        RPMprco(name=name, flags=flags, str_flags=self.RPM_PRCO_FLAGS_MAP[flags & 0xf], version=self._stringToVersion(version)))
        except:
            pass

        # obsoletes
        try:
            if self.header[1090]:
                for name, flags, version in zip(self.header[1090], self.header[1114], self.header[1115]):
                    self.obsoletes.append(
                        RPMprco(name=name, flags=flags, str_flags=self.RPM_PRCO_FLAGS_MAP[flags & 0xf], version=self._stringToVersion(version)))
        except:
            pass

        # conflicts
        try:
            if self.header[1054]:
                for name, flags, version in zip(self.header[1054], self.header[1053], self.header[1055]):
                    self.conflicts.append(
                        RPMprco(name=name, flags=flags, str_flags=self.RPM_PRCO_FLAGS_MAP[flags & 0xf], version=self._stringToVersion(version)))
        except:
            pass

    def _compute_checksum(self):
        self.rpmfile.seek(0)
        m = hashlib.sha256()
        size = 0
        data = self.rpmfile.read()
        while data:
            size += len(data)
            m.update(data)
            data = self.rpmfile.read()
        self.filesize = size
        self.checksum = m.hexdigest()

    def _stringToVersion(self, verstring):
        if verstring in [None, '']:
            return None, None, None
        i = verstring.find(':')
        if i != -1:
            try:
                epoch = str(int(verstring[:i]))
            except ValueError:
                # look, garbage in the epoch field, how fun, kill it
                epoch = '0'  # this is our fallback, deal
        else:
            epoch = '0'
        j = verstring.find('-')
        if j != -1:
            if verstring[i + 1:j] == '':
                version = None
            else:
                version = verstring[i + 1:j]
            release = verstring[j + 1:]
        else:
            if verstring[i + 1:] == '':
                version = None
            else:
                version = verstring[i + 1:]
            release = None
        return epoch, version, release



# -*- coding=utf-8 -*-
"""
Emulate a Maven3-compatible repository.


/$groupId[0]/../${groupId[n]/$artifactId/$version/$artifactId-$version.$extension
/$groupId[0]/../$groupId[n]/$artifactId/$version/$artifactId-$version-$classifier.$extension
/org/codehaus/plexus/plexus-container/0.15/
                     /plexus-container-0.15.jar
                     /plexus-container-0.15.jar.md5
                     /plexus-container-0.15.pom
                     /plexus-container-0.15.pom.md5
                     /plexus-container-0.15-javadoc.jar
                     /plexus-container-0.15-javadoc.jar.md5
                     /plexus-container-0.15-javasrc.jar
                     /plexus-container-0.15-javasrc.jar.md5

Maven3 variables:
    * $groupId = org.codehaus.plexus
    * $artifactId = plexus-container
    * $version = 0.15
    * $extension = jar
    * $classifier = javadoc, javasrc

Moneta mapping:
    * element.version maps to $version
    * element.archive maps to $groupId.$artifactId
    * element.name maps to $artifactId => built from element.archive
    * element.filename maps to $artifactId-$version-$classifier.$extension
    * element.full_name = element.filename

Command for uploading files:
curl --data-binary @$FILENAME http://[...]/\?filename=$FILENAME\&states=prod\&states=qualif\&archive=$groupId.$artifactId\&version=$version

URLs to emulate:
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename)$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).md5$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).sha1$
    /$groupId[0]/../$groupId[n]/$artifactId/$version/(filename).sha256$
"""
from django.core.urlresolvers import reverse

from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils.translation import ugettext as _

from moneta.repositories.maven3 import Maven3
from moneta.repository.models import Repository, ArchiveState, Element
from moneta.templatetags.moneta import moneta_url

__author__ = 'flanker'


class FlatFile(Maven3):
    verbose_name = _('Simple repository for flat files')
    storage_uid = '91439a10-0000-0000-0000-%012d'
    archive_type = 'flat_files'

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        if element.archive:
            element.name = element.archive.rpartition('.')[2]

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        Remove previous versions from target states
        :param element: Element
        :param states: list of ArchiveState
        """
        # remove previous versions from the given states:
        # noinspection PyUnresolvedReferences
        Element.states.through.objects.exclude(element__version=element.version) \
            .filter(archivestate__in=states, element__archive=element.archive).delete()
        super().finish_element(element, states)

    def index(self, request, rid):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = list(ArchiveState.objects.filter(repository=repo).order_by('name'))
        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'admin_allowed': repo.admin_allowed(request), }
        state_infos = []
        template_values['state_slug'] = None
        request_context = RequestContext(request)
        viewname = moneta_url(repo, 'browse')
        url = reverse(viewname, kwargs={'rid': repo.id, 'repo_slug': repo.slug, })

        state_infos.append(('all-packages', url, _('All states'), states))
        for state in states:
            template_values['state_slug'] = state.slug
            url = reverse(viewname, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'state_slug': state.slug})
            state_infos.append((state.slug, url, state.name, [state]))
        template_values['state_infos'] = state_infos
        return render_to_response('repositories/flat_files/index.html', template_values, request_context)



# coding=utf-8
import logging
import os
import uuid
import functools

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models import Q
from django.db.models.signals import pre_delete, post_migrate
from django.dispatch import receiver
from django.utils.text import slugify
from django.utils.translation import ugettext as _, ugettext

from moneta.exceptions import InvalidRepositoryException
from moneta.repository.storages import BaseStorage
from moneta.utils import normalize_str, remove, import_path

__author__ = 'flanker'


logger = logging.getLogger('moneta.files')


ARCHIVE_FILTER_CALLABLES = None
STORAGES = {}


@functools.lru_cache()
def storage(name) -> BaseStorage:
    kwargs = settings.STORAGES.get(name, settings.STORAGES['default'])
    cls = import_path(kwargs['ENGINE'])
    return cls(**kwargs)


@functools.lru_cache()
def archive_filters():
    """
    Check if archive filters set in settings are loaded. If not, load them.

    :return: a list of callable to call on new archives files.
    """
    result = []
    for middleware_path in settings.ARCHIVE_FILTERS:
        result.append(import_path(middleware_path))
    return result


class BaseModel(models.Model):
    name = models.CharField(_('Name'), max_length=100, blank=False, db_index=True)
    slug = models.SlugField(_('Slug'), max_length=100, blank=False, db_index=True)
    creation = models.DateTimeField(_('Creation date'), db_index=True, auto_now_add=True)
    modification = models.DateTimeField(_('Modification date'), db_index=True, auto_now=True)
    author = models.ForeignKey(settings.AUTH_USER_MODEL, verbose_name=_('Author'), db_index=True, blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super().save(*args, **kwargs)


class Repository(BaseModel):
    archive_type = models.CharField(_('Repository type'), db_index=True, max_length=100)
    on_index = models.BooleanField(_('Display on public index?'), db_index=True, default=True, blank=True)
    is_private = models.BooleanField(_('Should readers be authenticated?'), db_index=True, default=False, blank=True)
    admin_group = models.ManyToManyField(Group, verbose_name=_('Admin groups'), db_index=True, blank=True,
                                         related_name='repository_admin')
    reader_group = models.ManyToManyField(Group, verbose_name=_('Readers groups'), db_index=True, blank=True,
                                          related_name='repository_reader')

    def get_model(self):
        from moneta.repositories.base import RepositoryModelsClasses
        return RepositoryModelsClasses.get_model(self.archive_type)

    def get_absolute_url(self):
        return reverse('%s:index' % self.archive_type, kwargs={'rid': self.id})

    @staticmethod
    def index_queryset(request):
        user = request.user
        if user.is_anonymous():
            return Repository.objects.filter(Q(on_index=True) | Q(author=None)).distinct()
        return Repository.objects.filter(Q(on_index=True) | Q(author=user)).distinct()

    @staticmethod
    def upload_queryset(request):
        user = request.user
        if user.is_anonymous():
            return Repository.objects.filter(author=None)
        elif user.is_superuser:
            return Repository.objects.all()
        return Repository.objects.filter(Q(admin_group=user.groups.all()) | Q(author=user)).distinct()

    @staticmethod
    def admin_queryset(request):
        user = request.user
        if user.is_anonymous():
            return Repository.objects.filter(author=None)
        elif user.is_superuser:
            return Repository.objects.all()
        return Repository.objects.filter(author=user)

    @staticmethod
    def reader_queryset(request):
        user = request.user
        if user.is_anonymous():
            return Repository.objects.filter(Q(is_private=False) | Q(author=None)).distinct()
        return Repository.objects.filter(Q(is_private=False) | Q(author=user) | Q(reader_group=user.groups.all())
                                         | Q(admin_group=user.groups.all())).distinct()

    def upload_allowed(self, request):
        return self.upload_queryset(request).filter(id=self.id)[0:1].count() > 0

    def admin_allowed(self, request):
        return self.admin_queryset(request).filter(id=self.id)[0:1].count() > 0

    def reader_allowed(self, request):
        user = request.user
        if user.is_anonymous():
            return not self.is_private
        return not self.is_private or self.author == user or \
            (not {x.id for x in user.groups.all()}.isdisjoint({x.id for x in self.reader_group.all()})) or \
            (not {x.id for x in user.groups.all()}.isdisjoint({x.id for x in self.admin_group.all()}))


class ArchiveState(BaseModel):
    repository = models.ForeignKey(Repository, verbose_name=_('Repository'), db_index=True)


class Element(BaseModel):
    official_link = models.URLField(_('URL to the web page'), max_length=255, blank=True)
    short_description = models.CharField(_('Short description'), max_length=500, blank=True)
    long_description = models.TextField(_('Long description'), blank=True)
    states = models.ManyToManyField(ArchiveState, verbose_name=_('Archive states'), db_index=True)
    repository = models.ForeignKey(Repository, verbose_name=_('Repository'), db_index=True)
    full_name = models.CharField(_('Complete name'), max_length=255, blank=True, db_index=True)
    full_name_normalized = models.CharField(_('Normalized complete name'), max_length=255, blank=True, db_index=True,
                                            help_text=_("Complete name without special chars nor accents"))
    archive = models.CharField(_('Archive'), db_index=True, blank=True, default='', max_length=255)
    version = models.CharField(_('Version'), db_index=True, blank=True, default='', max_length=255)
    filename = models.CharField(_('filename'), max_length=255, blank=True, db_index=True, default='',
                                help_text=_('Automatically generated on creation'))
    uuid = models.CharField(_('UUID'), db_index=True, max_length=40, blank=True,
                            help_text=_('Unique identifier, automatically generated on first save'))
    md5 = models.CharField(_('MD5 sum'), max_length=120, blank=True, db_index=True, default='',
                           help_text=_('Automatically generated on creation'))
    sha1 = models.CharField(_('SHA1 sum'), max_length=120, blank=True, db_index=True, default='',
                            help_text=_('Automatically generated on creation'))
    sha256 = models.CharField(_('SHA256 sum'), max_length=120, blank=True, db_index=True, default='',
                              help_text=_('Automatically generated on creation'))
    filesize = models.IntegerField(_('File size'), default=0, blank=True, db_index=True,
                                   help_text=_('Automatically generated on creation'))
    extension = models.CharField(_('File extension'), db_index=True, max_length=20, blank=True, default='',
                                 help_text=_('Automatically generated on creation'))
    mimetype = models.CharField(_('MIME type'), db_index=True, max_length=40, blank=True, default='',
                                help_text=_('Guessed on creation'))
    archive_file = models.FileField(_('Archive file'), blank=True, default='', null=True,
                                    upload_to=settings.UPLOAD_ROOT, help_text=_('Any text, binary or archive file.'))
    archive_key = models.CharField(_('Original file'), blank=True, max_length=255, db_index=True, default='')
    uncompressed_key = models.CharField(_('Stored path'), blank=True, max_length=255, db_index=True, default='')
    extra_data = models.TextField(_('Extra repo data'), blank=True, default='')

    class Meta:
        verbose_name = _('file')
        verbose_name_plural = _('files')

    @staticmethod
    def reader_queryset(request):
        return Element.objects.filter(repository__in=Repository.reader_queryset(request))

    def remove_file(self):
        if self.archive_key:
            storage(settings.STORAGE_ARCHIVE).delete(self.archive_key)
            self.archive_key = None
        if self.uncompressed_key:
            storage(settings.STORAGE_UNCOMPRESSED).delete(self.uncompressed_key)
            self.uncompressed_key = None

    def delete(self, using=None):
        self.remove_file()
        result = super().delete(using=using)
        return result

    def save(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        if not self.uuid:
            self.uuid = str(uuid.uuid1())
        if (self.archive_file is not None) and self.archive_file.name:  # a new file is uploaded
            uploaded_file = self.archive_file
            # noinspection PyBroadException
            try:
                uploaded_file.open('rb')
                filename = os.path.basename(uploaded_file.name)
                self.set_file(uploaded_file.file, filename)
            except Exception:
                logger.error(ugettext('Unable to add the archive file'), exc_info=True)
            exceptions = []
            cls = self.repository.get_model()
            if not cls.is_file_valid(uploaded_file):
                exceptions.append(str(self.repository))
            else:
                cls.update_element(self)
            if len(exceptions) == 1:
                raise InvalidRepositoryException(
                    _('Repository %(repo)s is unable to handle this file.') % {'repo': exceptions[0], })
            elif len(exceptions) > 1:
                raise InvalidRepositoryException(
                    _('Repositories %(repo)s are unable to handle this file.') % {'repo': _(', ').join(exceptions), })
            self.archive_file = None
        self.full_name = self.filename
        self.full_name_normalized = normalize_str(self.full_name)
        super().save(*args, **kwargs)

    def set_file(self, obj_file, filename):
        temp_files = set()
        # noinspection PyBroadException
        try:
            self.remove_file()
            self.archive_key = storage(settings.STORAGE_ARCHIVE).store_descriptor(self.uuid, filename, obj_file)
            uncompressed_path = None
            for mw in archive_filters():
                obj_file.seek(0)
                u_path = mw(self, obj_file, filename, temp_files, uncompressed_path)
                if u_path is not None:
                    uncompressed_path = u_path
            if uncompressed_path is not None:  # there are extracted data to store
                self.uncompressed_key = storage(settings.STORAGE_UNCOMPRESSED).store(self.uuid, uncompressed_path)
        except Exception as e:
            logging.error(ugettext('Unable to add the archive file'), exc_info=True)
            raise e
        obj_file.seek(0)
        for f in temp_files:
            remove(f)

    def __str__(self):
        return self.full_name

    def get_absolute_url(self):
        return reverse('moneta.core.views.show_file', kwargs={'eid': self.id}, )

    def get_direct_link(self):
        return reverse('moneta.core.views.get_file', kwargs={'eid': self.id}, ) + '/' + self.filename


class ElementSignature(models.Model):
    GPG = 'gpg'
    OPENSSL = 'openssl'
    METHODS = ((GPG, _('GnuPG')), (OPENSSL, _('OpenSSL/x509')))
    element = models.ForeignKey(Element, verbose_name=_('Element'), db_index=True, default=0)
    signature = models.TextField(_('signature'), default='', blank=True)
    method = models.CharField(_('signature method'), choices=METHODS, db_index=True, max_length=10)
    creation = models.DateTimeField(_('Creation date'), db_index=True, auto_now_add=True)


# noinspection PyUnusedLocal
@receiver(pre_delete, sender=Element)
def delete_element(sender, instance=None, **kwargs):
    """


    :type sender: object
    :param sender:
    :param instance:
    :param kwargs:
    """
    if instance is not None and isinstance(instance, Element):
        instance.remove_file()


@receiver(post_migrate)
def create_default_group(signal, sender, app_config, verbosity, interactive, using, **kwargs):
    if sender.name != 'moneta.repository':
        return
    group, created = Group.objects.get_or_create(name=settings.FLOOR_DEFAULT_GROUP_NAME)
    perms = list(Permission.objects.filter(codename='add_repository')[0:1])
    if perms:
        group.permissions.add(*perms)
